import pathlib

from escudeiro import autodiscovery
from escudeiro.misc import filter_isinstance, filter_issubclass


def test_runtime_autodiscovery_childof(
    sample_root_path: pathlib.Path, mock_module_path: pathlib.Path
):
    Test = __import__(
        "tests.autodiscovery.modules.mod_a", fromlist=["Test"]
    ).Test
    runtime = autodiscovery.RuntimeAutoDiscovery(
        autodiscovery.runtime_child_of(Test),
        sample_root_path,
        mock_module_path,
    )
    result = dict(runtime)

    assert result
    assert list(filter_issubclass(Test, result.values()))


def test_runtime_autodiscovery_instanceof(
    sample_root_path: pathlib.Path, mock_module_path: pathlib.Path
):
    Test = __import__(
        "tests.autodiscovery.modules.mod_a", fromlist=["Test"]
    ).Test
    runtime = autodiscovery.RuntimeAutoDiscovery(
        autodiscovery.runtime_instance_of(Test),
        sample_root_path,
        mock_module_path,
    )
    result = dict(runtime)

    assert result
    assert list(filter_isinstance(Test, result.values()))


def test_runtime_autodiscovery_contains_attr(
    sample_root_path: pathlib.Path, mock_module_path: pathlib.Path
):
    runtime = autodiscovery.RuntimeAutoDiscovery(
        autodiscovery.runtime_contains_attr("name"),
        sample_root_path,
        mock_module_path,
    )
    result = dict(runtime)

    assert result
    assert all(hasattr(item, "name") for item in result.values())


def test_runtime_autodiscovery_attr_with_value(
    sample_root_path: pathlib.Path, mock_module_path: pathlib.Path
):
    runtime = autodiscovery.RuntimeAutoDiscovery(
        autodiscovery.runtime_attr_with_value("abstract", False),
        sample_root_path,
        mock_module_path,
    )
    result = dict(runtime)

    assert result
    assert all(
        hasattr(item, "abstract") and item.abstract is False
        for item in result.values()
    )



from typing import Annotated, Any, Literal, override

import pytest

from escudeiro.misc.typex import (
    assert_notnone,
    cast_notnone,
    is_hashable,
    is_instanceexact,
)


class TestIsHashable:
    # ==== Types that SHOULD be hashable ====
    @pytest.mark.parametrize(
        "typ",
        [
            int,
            str,
            float,
            bool,
            type(None),
            int | None,
            int | str,
            tuple[int, str],
            Literal[1, 2, 3],
            Annotated[int, "metadata"],
            frozenset[str],
        ],
    )
    def test_hashable_types(self, typ: Any):
        assert is_hashable(typ), f"{typ} should be hashable"

    # ==== Types that SHOULD NOT be hashable ====
    @pytest.mark.parametrize(
        "typ",
        [
            list[int],
            set[str],
            dict[str, int],
            list[str] | None,
            str | list[str],
            tuple[int, list[int]],
            Annotated[list[int], "metadata"],
        ],
    )
    def test_unhashable_types(self, typ: Any):
        assert not is_hashable(typ), f"{typ} should NOT be hashable"

    # ==== Edge: recursive or nested generics ====
    @pytest.mark.parametrize(
        "typ",
        [
            int | list[str] | None,
            tuple[int, int] | list[int],
            Annotated[dict[str, int] | None, "meta"],
        ],
    )
    def test_complex_unhashable_cases(self, typ: Any):
        assert not is_hashable(typ), (
            f"{typ} should NOT be hashable (deep unhashable part)"
        )

    # ==== User-defined classes ====

    class HashableCustom:
        @override
        def __hash__(self):
            return 42

    class UnhashableCustom:
        __hash__ = None  # pyright: ignore[reportAssignmentType]

    def test_custom_class_hashable(self):
        assert is_hashable(self.HashableCustom)

    def test_custom_class_unhashable(self):
        assert not is_hashable(self.UnhashableCustom)

    # ==== Aliased types / runtime aliases ====
    MyFrozenSet = frozenset[int]
    MyList = list[str]
    type MyTuple = tuple[int]
    type MyDict = dict[str, Any]

    def test_alias_variable(self):
        assert is_hashable(self.MyFrozenSet)
        assert not is_hashable(self.MyList)

    def test_type_alias_type(self):
        assert is_hashable(self.MyTuple)
        assert not is_hashable(self.MyDict)

    # ==== Optional Ellipsis corner case ====
    def test_ellipsis_is_ignored(self):
        assert is_hashable(None | tuple[int, ...])


class TestIsInstanceexact:
    @pytest.mark.parametrize(
        "obj, annotation, expected",
        [
            (1, int, True),
            (1.0, float, True),
            ("test", str, True),
            (None, type(None), True),
            (1, int | str, True),
            (1.0, int | str, False),
            ("test", int | str, True),
            ([1, 2], list[int], True),
            (frozenset([1]), frozenset[int], True),
        ],
        ids=[
            "int_exact",
            "float_exact",
            "str_exact",
            "none_exact",
            "int_or_str_match_int",
            "int_or_str_mismatch_float",
            "int_or_str_match_str",
            "list_not_instanceexact",
            "frozenset_exact",
        ],
    )
    def test_is_instanceexact(self, obj: Any, annotation: Any, expected: bool):
        assert is_instanceexact(obj, annotation) == expected

    def test_does_not_match_for_subclasses(self):
        class Base:
            pass

        class Derived(Base):
            pass

        assert is_instanceexact(Derived(), Base) is False
        assert is_instanceexact(Base(), Base) is True
        assert is_instanceexact(Derived(), Derived) is True

    def test_union_type_handling(self):
        class CustomType:
            pass

        assert is_instanceexact(CustomType(), CustomType | int) is True
        assert is_instanceexact(1, CustomType | int) is True
        assert is_instanceexact("test", CustomType | int) is False


class TestAssertNotNone:
    def test_assert_notnone_with_value(self):
        assert assert_notnone(5) == 5
        assert assert_notnone("hello") == "hello"

    def test_assert_notnone_with_none(self):
        with pytest.raises(ValueError, match="Value is None"):
            assert_notnone(None)


class TestCastNotNone:
    def test_cast_notnone_with_value(self):
        assert cast_notnone(5) == 5
        assert cast_notnone("hello") == "hello"

    def test_cast_notnone_with_none(self):
        assert cast_notnone(None) is None, "cast_notnone does not raise, just casts"
    



from escudeiro.url.query import Query


def test_query():  # sourcery skip: extract-duplicate-method
    # Test query load
    query = Query("param1=value1&param2=value2")
    assert query.params == {"param1": ["value1"], "param2": ["value2"]}

    # Test query encode
    query = Query("param1=value1&param2=value2")
    assert query.encode() == "param1=value1&param2=value2"

    # Test query no null equal returns empty param
    query = Query("param1=&param2=value2")
    assert query.omit_empty_equal() == "param1&param2=value2"

    # Test query adds correctly
    query = Query("param1=value1&param2=value2").add(param3="value3")
    assert query.params == {
        "param1": ["value1"],
        "param2": ["value2"],
        "param3": ["value3"],
    }

    # Test query set adds correctly
    query = Query("param1=value1&param2=value2").set(param3="value3")
    assert query.params == {"param3": ["value3"]}

    # Test setitem works correctly
    query = Query("param1=value1&param2=value2")
    query["param3"] = "value3"
    assert query.params == {
        "param1": ["value1"],
        "param2": ["value2"],
        "param3": ["value3"],
    }

    # Test getitem works correctly
    query = Query("param1=value1&param2=value2")
    assert query["param1"] == ["value1"]

    # Test removing a single parameter
    q = Query("param1=value1&param2=value2").remove("param1")
    assert q.encode() == "param2=value2"

    # Test removing multiple parameters at once
    q = Query("param1=value1&param2=value2&param3=value3").remove(
        "param1", "param3"
    )
    assert q.encode() == "param2=value2"

    # Test removing a non-existent parameter
    q = Query("param1=value1").remove("param2")
    assert q.encode() == "param1=value1"

    # Test Query properly encodes slashes
    q = Query("param1=va/lue").add({"param2": "val/ie"})

    assert q.encode() == "param1=va%2Flue&param2=val%2Fie"



from escudeiro.url.path import Path


def test_normalize_path():
    # Test case 1: Basic usage
    assert Path("/a//./b/..//c/").normalize().encode() == "/a/c/"

    # Test case 2: Handling segments starting with '.'
    assert Path("//a/.b///c//").normalize().encode() == "/a/.b/c/"

    # Test case 3: Handling multiple ".." and "." segments
    assert (
        Path("/a/b/c/d/e/f/../g/./h/i/j/").normalize().encode()
        == "/a/b/c/d/e/g/h/i/j/"
    )

    # Test case 4: Handling an empty path
    assert Path("").normalize().encode() == ""

    # Test case 5: Handling a single "/"
    assert Path("/").normalize().encode() == "/"

    # Test case 6: Handling multiple "/"
    assert Path("////a/b/c/d").normalize().encode() == "/a/b/c/d"

    # Test case 7: Handling multiple segments starting with '.'
    assert Path("/.b/.c/.d/e/").normalize().encode() == "/.b/.c/.d/e/"


def test_path():
    # Test basic initialization
    path = Path("/path/to/file")
    assert path.segments == ["", "path", "to", "file"]

    # Test encoding
    assert path.encode() == "/path/to/file"

    # Test adding a path
    _ = path.add("/another/path")
    assert path.segments == ["", "path", "to", "file", "another", "path"]
    assert path.encode() == "/path/to/file/another/path"

    # Test setting a path
    _ = path.set("/new/path")
    assert path.segments == ["", "new", "path"]
    assert path.encode() == "/new/path"

    # Test isdir property
    path = Path("/path/to/")
    assert path.isdir
    path = Path("/path/to/file")
    assert not path.isdir

    # Test isfile property
    path = Path("/path/to/")
    assert not path.isfile
    path = Path("/path/to/file")
    assert path.isfile

    # Test normalize
    path = Path("/path/to/file/../another/path/").normalize()
    assert path.segments == ["", "path", "to", "another", "path", ""]
    assert path.encode() == "/path/to/another/path/"

    # Test empty str on Path
    path = Path("")

    assert path.segments == []
    assert path.encode() == ""

    # Test with a single forward slash
    path = Path("/")

    assert path.segments == ["", ""]
    assert path.encode() == "/"

    # Test with a lot of slashes
    path = Path("////a////b//")

    assert path.segments == ["", "a", "", "", "", "b", ""]

    # Test if path encodes correctly
    path = Path("/a b/c d/")

    assert path.encode() == "/a%20b/c%20d/"

    # Test if path does not overencodes string
    path = Path("/a%20b/c%20d/")

    assert path.encode() == "/a%20b/c%20d/"



"""Utilities for safer type casting and function execution.

This module provides functions for safely performing type casting operations
and controlling function execution patterns.
"""

from __future__ import annotations

import asyncio
import functools
import re
import typing
from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    Awaitable,
    Callable,
    Collection,
    Coroutine,
    Generator,
    Iterable,
)
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    contextmanager,
)
from dataclasses import dataclass
from datetime import date, datetime, time, tzinfo
from time import sleep
from types import FunctionType, LambdaType, MethodType
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Literal,
    cast,
    final,
    overload,
)

from escudeiro.exc.errors import ErrorGroup, InvalidCast, RetryError
from escudeiro.misc.contextx import AsyncContextWrapper, is_async_context
from escudeiro.misc.strings import sentence, squote


def safe_cast[V, T, D](
    caster: Callable[[V], T],
    value: V,
    *ignore_childof: type[Exception],
    default: D = None,
) -> T | D:
    """Safely casts a value using the provided casting function.

    Attempts to cast the given value using the casting function. If the casting
    operation raises any of the specified exceptions, returns the default value.

    Args:
        caster: A function that converts the input value to the target type.
        value: The value to be cast.
        *ignore_childof: Exception types to catch. Defaults to (TypeError, ValueError)
                         if none are specified.
        default: The value to return if casting fails. Defaults to None.

    Returns:
        The result of the casting operation if successful, otherwise the default value.
    """
    if not ignore_childof:
        ignore_childof = (TypeError, ValueError)

    try:
        return caster(value)
    except ignore_childof:
        return default


async def asafe_cast[V, D, T](
    caster: Callable[[V], Awaitable[T]],
    value: V,
    *ignore_childof: type[Exception],
    default: D = None,
) -> T | D:
    """Safely casts a value using the provided asynchronous casting function.

    Asynchronous version of safe_cast. Attempts to cast the given value using the
    asynchronous casting function. If the casting operation raises any of the
    specified exceptions, returns the default value.

    Args:
        caster: An asynchronous function that converts the input value to the target type.
        value: The value to be cast.
        *ignore_childof: Exception types to catch. Defaults to (TypeError, ValueError)
                         if none are specified.
        default: The value to return if casting fails. Defaults to None.

    Returns:
        The result of the casting operation if successful, otherwise the default value.
    """
    if not ignore_childof:
        ignore_childof = (TypeError, ValueError)
    try:
        return await caster(value)
    except ignore_childof:
        return default


def call_once[T](func: Callable[[], T]) -> Callable[[], T]:
    """Returns a wrapper that ensures the wrapped function is called only once.

    Creates a wrapper around the provided function that caches the result of the
    first call. Subsequent calls return the cached result without executing
    the original function again.

    Args:
        func: The function to be wrapped. Must take no arguments.

    Returns:
        A wrapped version of the function that executes at most once.
    """
    sentinel: Any = object()
    output = sentinel

    @functools.wraps(func)
    def wrapper() -> T:
        nonlocal output
        if output is sentinel:
            output = func()
        return output

    return wrapper


type AsyncFunc[**P, T] = Callable[P, Coroutine[Any, Any, T]]
type AsAsyncFactory[**P, T] = Callable[
    Concatenate[Callable[P, T], P], Coroutine[Any, Any, T]
]

if TYPE_CHECKING:
    asyncio.to_thread = typing.cast(AsAsyncFactory, asyncio.to_thread)


@overload
def as_async[**P, T](
    func: None = None,
    /,
    *,
    cast: AsAsyncFactory[P, T] = asyncio.to_thread,
) -> Callable[[Callable[P, T]], AsyncFunc[P, T]]: ...


@overload
def as_async[**P, T](
    func: AsyncFunc[P, T],
    /,
) -> AsyncFunc[P, T]: ...


@overload
def as_async[**P, T](
    func: Callable[P, T],
    /,
) -> AsyncFunc[P, T]: ...


def as_async[**P, T](
    func: Callable[P, T] | AsyncFunc[P, T] | None = None,
    /,
    *,
    cast: AsAsyncFactory[P, T] = asyncio.to_thread,
) -> AsyncFunc[P, T] | Callable[[Callable[P, T]], AsyncFunc[P, T]]:
    """Convert a synchronous function to an asynchronous one.

    If the function is already asynchronous, returns it unchanged.
    Otherwise, wraps it using the specified casting method (defaults to asyncio.to_thread).

    Can be used as a decorator with or without arguments.

    Args:
        func: The function to convert to asynchronous. If None, returns a decorator.
        cast: A factory function that converts a synchronous function to asynchronous.
            Defaults to asyncio.to_thread, which runs the function in a separate thread.

    Returns:
        An asynchronous version of the input function, or a decorator that produces one.

    Examples:
        ```python
        # Basic usage as a decorator
        @as_async
        def read_file(path):
            with open(path) as f:
                return f.read()

        # With custom executor
        @as_async(cast=my_custom_executor)
        def process_data(data):
            return expensive_calculation(data)

        # Direct usage
        async_read = as_async(read_file)
        ```
    """

    def outer(func: Callable[P, T] | AsyncFunc[P, T]) -> AsyncFunc[P, T]:
        if asyncio.iscoroutinefunction(func):
            return func
        elif TYPE_CHECKING:
            func = typing.cast(Callable[P, T], func)

        @functools.wraps(func)
        async def _inner(*args: P.args, **kwargs: P.kwargs) -> T:
            return await cast(func, *args, **kwargs)

        return _inner

    if func is None:
        return outer
    return outer(func)


def cache[**P, T](f: Callable[P, T]) -> Callable[P, T]:
    """Simple wrapper around functools.cache to preserve function signature.

    Provides memoization for the wrapped function, caching return values based on arguments.
    This is a thin wrapper over functools.cache that preserves proper type annotations.

    Args:
        f: The function to cache.

    Returns:
        A memoized version of the function with proper type annotations.

    Examples:
        ```python
        @cache
        def fibonacci(n):
            if n <= 1:
                return n
            return fibonacci(n-1) + fibonacci(n-2)
        ```
    """
    return typing.cast(Callable[P, T], functools.cache(f))


@overload
def make_noop[T](
    *, returns: T = None, asyncio: Literal[False] = False
) -> Callable[..., T]: ...


@overload
def make_noop[T](
    *, returns: T = None, asyncio: Literal[True]
) -> Callable[..., Coroutine[Any, Any, T]]: ...


def make_noop[T](
    *, returns: T = None, asyncio: bool = False
) -> Callable[..., T] | Callable[..., Coroutine[Any, Any, T]]:
    """Creates a no-operation function that accepts any arguments and returns a fixed value.

        Useful for creating placeholders, stubs, or disabling functionality temporarily.

        Args:
            returns: The value the function should return. Defaults to None.
            asyncio: Whether to create an async function. Defaults to False.

        Returns:
            A function that ignores all arguments and returns the specified value,
            either synchronously or asynchronously based on the asyncio parameter.

        Examples:
            ```python
            # Synchronous no-op
    @overload
    def make_noop[T](
        *, returns: T = None, asyncio: Literal[True]
    ) -> Callable[..., Coroutine[Any, Any, T]]: ...

            log_function = make_noop()

            # No-op with custom return value
            get_user = make_noop(returns={"id": 0, "name": "Guest"})

            # Async no-op
            async_operation = make_noop(asyncio=True, returns={"status": "success"})
            ```
    """

    def _noop(*args: Any, **kwargs: Any) -> T:
        del args, kwargs
        return returns

    return _noop if not asyncio else as_async(_noop)


def return_param[T](param: T) -> T:
    """Returns the provided parameter without modification.

    This function is a simple utility that returns the input parameter as-is.
    It can be useful in functional programming patterns where you need to pass
    a function that simply returns its input.

    Args:
        param: The parameter to return.

    Returns:
        The same parameter that was passed in.

    Example:
        ```python
        value = return_param(42)  # Returns 42
        ```
    """
    return param


def do_with[**P, T](
    context_manager: AbstractContextManager,
    func: Callable[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    """Executes a function within the context of a context manager.

    Allows for functional-style usage of context managers by automatically
    entering the context, executing the function, and exiting the context.

    Args:
        context_manager: The context manager to use.
        func: The function to execute within the context.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The result of calling the function with the specified arguments.

    Examples:
        ```python
        # Execute in a transaction
        result = do_with(
            db.transaction(),
            create_user,
            username="johndoe",
            email="john@example.com"
        )
        ```
    """
    with context_manager:
        return func(*args, **kwargs)


async def asyncdo_with[**P, T](
    context_manager: AbstractContextManager | AbstractAsyncContextManager,
    func: Callable[P, T] | AsyncFunc[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    """Executes a function within an async or sync context manager.

    Asynchronous version of do_with. Works with both synchronous and
    asynchronous context managers and functions.

    Args:
        context_manager: The context manager to use. Can be synchronous or asynchronous.
        func: The function to execute within the context. Can be synchronous or asynchronous.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The result of calling the function with the specified arguments.

    Examples:
        ```python
        # With async context manager and function
        result = await asyncdo_with(
            async_db.transaction(),
            async_create_user,
            username="johndoe"
        )

        # With sync context manager and async function
        result = await asyncdo_with(
            lock_manager(),
            async_process_data,
            data=payload
        )
        ```
    """
    if not is_async_context(context_manager):
        context_manager = AsyncContextWrapper(context_manager)

    async with context_manager:
        return await as_async(func)(*args, **kwargs)  # pyright: ignore[reportReturnType]


@final
class FrozenCoroutine[T]:
    """A coroutine wrapper that ensures the wrapped coroutine is executed at most once.

    When awaited multiple times, the coroutine is executed only on the first await.
    Subsequent awaits return the cached result without re-executing the coroutine.
    Thread-safe through the use of an asyncio.Lock.

    Attributes:
        _coro: The wrapped coroutine.
        _lock: Lock to ensure thread safety.
        _value: Cached result after first execution, or sentinel if not yet executed.

    Examples:
        ```python
        # Direct usage
        async def fetch_data():
            print("Fetching data...")
            return await api.get_data()

        frozen = FrozenCoroutine(fetch_data())

        # First await executes the coroutine
        result1 = await frozen  # Prints "Fetching data..."

        # Second await returns cached result
        result2 = await frozen  # No print, returns same result
        ```
    """

    _sentinel: Any = object()

    def __init__(self, coro: Coroutine[Any, Any, T]) -> None:
        """Initialize with a coroutine to be executed at most once.

        Args:
            coro: The coroutine to wrap.
        """
        self._coro = coro
        self._lock = asyncio.Lock()
        self._value = self._sentinel

    async def _await(self) -> T:
        """Internal method that handles the actual awaiting logic.

        Ensures the coroutine is executed at most once and returns
        the cached result for subsequent calls.

        Returns:
            The result of the coroutine execution.
        """
        async with self._lock:
            if self._value is not self._sentinel:
                return self._value
            self._value = await self._coro
            return self._value

    def __await__(self):
        return self._await().__await__()

    @classmethod
    def decorate[**P](
        cls, func: Callable[P, Coroutine[Any, Any, T]]
    ) -> Callable[P, FrozenCoroutine[T]]:
        """Decorator that wraps a coroutine-returning function to ensure
        each returned coroutine is executed at most once.

        Args:
            func: A function that returns a coroutine.

        Returns:
            A function that returns a FrozenCoroutine wrapping the original
            function's coroutine.

        Examples:
            ```python
            @FrozenCoroutine.decorate
            async def fetch_user(user_id):
                print(f"Fetching user {user_id}")
                return await api.get_user(user_id)

            # Creates a FrozenCoroutine
            user_coro = fetch_user(123)

            # First await executes
            user1 = await user_coro  # Prints "Fetching user 123"

            # Second await returns cached result
            user2 = await user_coro  # No print, returns same result
            ```
        """

        @functools.wraps(func)
        def _wrapper(*args: P.args, **kwargs: P.kwargs) -> FrozenCoroutine[T]:
            return FrozenCoroutine(func(*args, **kwargs))

        return _wrapper


def as_datetime(value: date, tz: tzinfo | None = None) -> datetime:
    return datetime(value.year, value.month, value.day, tzinfo=tz)


def join_into_datetime(
    dt: date, tm: time, tz: tzinfo | None = None
) -> datetime:
    return as_datetime(dt, tz).replace(
        hour=tm.hour,
        minute=tm.minute,
        second=tm.second,
        microsecond=tm.microsecond,
    )


async def as_async_iterable[T](iterable: Iterable[T]) -> AsyncIterable[T]:
    for item in iterable:
        yield item


@contextmanager
def raise_insteadof(
    insteadof: type[Exception], exc: type[Exception] = Exception, *args: Any
) -> Generator[None, None, None]:
    """Replaces an exception with another within a context.

    This context manager catches exceptions of type `insteadof` and raises a new
    exception of type `exc` instead, optionally passing arguments to the new exception.

    Args:
        insteadof: The exception type to catch.
        exc: The exception type to raise instead. Defaults to `Exception`.
        *args: Arguments to pass to the new exception.

    Raises:
        exc: If `insteadof` is raised, it is replaced with an instance of `exc`.

    Example:
        ```python
        try:
            with raise_insteadof(KeyError, ValueError, "Invalid key!"):
                raise KeyError("Missing key")
        except ValueError as e:
            print(e)  # Outputs: "Invalid key!"
        ```
    """
    try:
        yield
    except insteadof:
        raise exc(*args) from None


@dataclass
class Retry:
    """
    A utility class that provides retry logic for functions.
    This class allows you to retry a function multiple times in case of an exception. It supports both synchronous
    and asynchronous functions. You can configure the number of retries, the delay between retries, and an optional
    backoff multiplier for exponential delay growth.
    Attributes:
        signal: The exception type(s) that triggers the retry logic.
        count: The number of retries to attempt before giving up. Default is 3.
        delay: The initial delay (in seconds) between retries. Default is 0.
        backoff: The multiplier for the delay between retries (exponential backoff). Default is 1 (no backoff).
    Example:
        ```python
        retry = Retry(signal=ConnectionError, count=5, delay=1, backoff=2)
        @retry
        def unreliable_function():
            # Function implementation that might raise ConnectionError
            pass
        ```
    """

    signal: type[Exception] | tuple[type[Exception], ...]
    count: int = 3
    delay: float = 0  # seconds between retries
    backoff: float = 1  # multiplier for delay

    def get_signal(self) -> tuple[type[Exception], ...]:
        return (
            self.signal if isinstance(self.signal, tuple) else (self.signal,)
        )

    def __call__[**P, T](self, func: Callable[P, T]) -> Callable[P, T]:
        """
        Decorator that retries a function call on failure.
        This method wraps the provided function, retrying it on failure for a specified number of attempts. If the
        function raises the exception specified by `signal`, it will retry the function call up to the `count`
        number of times, with a delay between each retry and optional exponential backoff.
        Args:
            func: The function to be wrapped with retry logic.
        Returns:
            A wrapped function that will retry on failure.
        Example:
            ```python
            @retry
            def fetch_data():
                # Function that might raise a ConnectionError
                pass
            ```
        """

        @functools.wraps(func)
        def retrier(*args: P.args, **kwargs: P.kwargs) -> T:
            fails = []
            current_delay = self.delay
            signal = self.get_signal()
            for attempt in range(self.count):
                try:
                    return func(*args, **kwargs)
                except signal as e:
                    fails.append(e)
                    if attempt < self.count - 1 and current_delay > 0:
                        sleep(current_delay)
                        current_delay *= self.backoff
            raise ErrorGroup(
                "Failed retry operation",
                (RetryError(f"Exceeded max retries: {self.count}"), *fails),
            )

        return retrier

    def acall[**P, T](
        self, func: Callable[P, Coroutine[Any, Any, T]]
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        """
        Asynchronous version of the retry decorator for functions that return coroutines.
        This method wraps the provided asynchronous function, retrying it on failure for a specified number of
        attempts. If the function raises the exception specified by `signal`, it will retry the function call
        up to the `count` number of times, with a delay between each retry and optional exponential backoff.
        Args:
            func: The asynchronous function to be wrapped with retry logic.
        Returns:
            A wrapped asynchronous function that will retry on failure.
        Example:
            ```python
            @retry.acall
            async def fetch_data():
                # Async function that might raise a ConnectionError
                pass
            ```
        """

        @functools.wraps(func)
        async def retrier(*args: P.args, **kwargs: P.kwargs) -> T:
            fails = []
            current_delay = self.delay
            signal = self.get_signal()
            for attempt in range(self.count):
                try:
                    return await func(*args, **kwargs)
                except signal as e:
                    fails.append(e)
                    if attempt < self.count - 1 and current_delay > 0:
                        await asyncio.sleep(current_delay)
                        current_delay *= self.backoff
            raise ErrorGroup(
                "Failed retry operation",
                (RetryError(f"Exceeded max retries: {self.count}"), *fails),
            )

        return retrier

    def map[S, T](
        self,
        predicate: Callable[[S], T],
        collection: Iterable[S],
        strategy: Literal["threshold", "temperature"] = "threshold",
    ) -> Iterable[T]:
        """
        Applies the retry logic to each item in a collection.
        This method attempts to apply the given predicate function to each item in the collection. If an exception
        occurs during the execution of the predicate, it will retry the operation up to the `count` number of times,
        with a delay between retries. You can control the retry strategy with the `strategy` parameter.
        Args:
            predicate: The function to apply to each item in the collection.
            collection: The collection of items to iterate over and apply the predicate to.
            strategy: The retry strategy to use. "threshold" retries a fixed number of times, while "temperature"
                      decreases the retry count with each successful operation. Default is "threshold".
        Returns:
            An iterable of the results of the predicate applied to each item in the collection.
        Example:
            ```python
            retry.map(fetch_data, [item1, item2])
            ```
        """
        count = 0
        predicate = self(predicate)
        signal = self.get_signal()
        for item in collection:
            try:
                yield predicate(item)
            except signal as err:
                if count >= self.count:
                    raise ErrorGroup(
                        "Failed retry operation",
                        (
                            RetryError(f"Exceeded max retries: {self.count}"),
                            err,
                        ),
                    ) from err
                count += 1
            else:
                if strategy == "temperature":
                    count = max(0, count - 1)

    async def amap[S, T](
        self,
        predicate: Callable[[S], Coroutine[Any, Any, T]],
        collection: Collection[S],
        strategy: Literal["threshold", "temperature"] = "threshold",
    ) -> AsyncIterable[T]:
        """
        Asynchronous version of `map` for a collection of items, using coroutines for the predicate.
        This method attempts to apply the given asynchronous predicate function to each item in the collection.
        If an exception occurs during the execution of the predicate, it will retry the operation up to the `count`
        number of times, with a delay between retries. The retry strategy can be controlled using the `strategy`
        parameter.
        Args:
            predicate: The asynchronous function to apply to each item in the collection.
            collection: The collection of items to iterate over and apply the predicate to.
            strategy: The retry strategy to use. "threshold" retries a fixed number of times, while "temperature"
                      decreases the retry count with each successful operation. Default is "threshold".
        Returns:
            An asynchronous iterable of the results of the predicate applied to each item in the collection.
        Example:
            ```python
            async for result in retry.amap(fetch_data, [item1, item2]):
                print(result)
            ```
        """
        count = 0
        signal = self.get_signal()
        predicate = self.acall(predicate)
        for item in collection:
            try:
                yield await predicate(item)
            except signal as err:
                if count >= self.count:
                    raise ErrorGroup(
                        "Failed retry operation",
                        (
                            RetryError(f"Exceeded max retries: {self.count}"),
                            err,
                        ),
                    ) from err
                count += 1
            else:
                if strategy == "temperature":
                    count = max(0, count - 1)

    async def agenmap[S, T](
        self,
        predicate: Callable[[S], Coroutine[Any, Any, T]],
        collection: AsyncGenerator[S],
        strategy: Literal["threshold", "temperature"] = "threshold",
    ) -> AsyncIterable[T]:
        """
        Asynchronously applies retry logic to each item generated by the given asynchronous generator.
        This method applies the given asynchronous predicate function to each item generated by the asynchronous
        generator. If an exception occurs during the execution of the predicate, it will retry the operation up to
        the `count` number of times, with a delay between retries. The retry strategy can be controlled using the
        `strategy` parameter.
        Args:
            predicate: The asynchronous function to apply to each item from the generator.
            collection: The asynchronous generator of items to iterate over and apply the predicate to.
            strategy: The retry strategy to use. "threshold" retries a fixed number of times, while "temperature"
                        decreases the retry count with each successful operation. Default is "threshold".
        Returns:
            An asynchronous iterable of the results of the predicate applied to each item from the generator.
        Example:
            ```python
            async for result in retry.agenmap(fetch_data, async_item_generator()):
                print(result)
            ```
        """
        count = 0
        signal = self.get_signal()
        predicate = self.acall(predicate)
        async for item in collection:
            try:
                yield await predicate(item)
            except signal as err:
                if count >= self.count:
                    raise ErrorGroup(
                        "Failed retry operation",
                        (
                            RetryError(f"Exceeded max retries: {self.count}"),
                            err,
                        ),
                    ) from err
                count += 1
            else:
                if strategy == "temperature":
                    count = max(0, count - 1)


UNSET = object()
_index_pattern = re.compile(r"\[([0-9]+)\]")


def walk_object(obj: Any, path: str) -> Any:
    """Safely retrieves a value from an object using a dot-separated path.
    This function allows you to access nested attributes, dictionary keys, or sequence indices
    in a safe manner, returning None if any part of the path is not found.
    Args:
        obj: The object to traverse.
    Returns:
        The value at the specified path, or None if the path does not exist.
    Example:
        ```python
        class User:
            def __init__(self, name, age):
                self.name = name
                self.age = age
        user = User("Alice", 30)
        age = walk_object(user, "age")  # Returns 30
        nested_dict = {"user": {"name": "Bob", "details": {"age": 25}}}
        name = walk_object(nested_dict, "user.name")  # Returns "Bob"
        invalid = walk_object(nested_dict, "user.details.address")  # Returns None
        nested_list = [1, 2, [3, 4, 5]]
        value = walk_object(nested_list, "[2].[1]")  # Returns 4
        ```
    """
    parts = path.split(".")
    value = obj
    for part in parts:
        if v := _index_pattern.match(part):
            idx = int(v.group(1))
            if TYPE_CHECKING:
                assert isinstance(value, list | tuple), (
                    "Expected a list or tuple"
                )
            value = value[idx] if len(value) > idx else UNSET
        elif isinstance(value, dict):
            value = value.get(part, UNSET)
        else:
            value = getattr(value, part, UNSET)
        if value in (None, UNSET):
            break
    return value if value is not UNSET else None


def isinstance_or_cast[T, Arg, U](
    expects: type[T], onmiss: Callable[[Arg], U]
) -> Callable[[Arg | T], T | U]:
    """
    Returns a decorator that checks if the value is an instance of the given type.
    If it is, it returns the value, otherwise it calls the given callable.
    """

    @functools.wraps(onmiss)
    def _check(value: Any) -> T | U:
        if isinstance(value, expects):
            return value
        return onmiss(value)

    return _check


class Caster[U, T]:
    def __init__(
        self, caster: Callable[[U], T], name: str | None = None
    ) -> None:
        self._caster = caster
        self._name = cast(
            str, name or getattr(caster, "__name__", type(self).__name__)
        )

    def __call__(self, value: U) -> T:
        """Casts the value using the provided caster function."""
        return self._caster(value)

    def __name__(self) -> str:
        return self._name

    def join[S](self, caster: Callable[[T], S]) -> Caster[U, S]:
        def _join(value: U) -> S:
            return caster(self._caster(value))

        return Caster(
            _join,
            name=f"{self._name}.join({getattr(caster, '__name__', type(caster).__name__)})",
        )

    cast = join  # Alias for join method for compatibility with _JoinedCast

    def strict(self, value: U) -> T:
        """Casts the value and raises an error if the result is None."""
        result = self._caster(value)
        if result is None:
            raise InvalidCast(
                sentence(
                    f"Received falsy value {result} from {value} during cast"
                ),
                result,
            )
        return result

    def safe[S](
        self,
        value: U,
        *childof: type[Exception],
        default: T | S = None,
    ) -> T | S:
        """Casts the value and returns None if the result is None, otherwise returns the result."""
        return safe_cast(
            self,
            value,
            *childof,
            default=default,
        )

    optional = (
        safe  # Alias for safe method for compatibility with maybe_result
    )

    def safe_cast(
        self,
        child_of: Collection[type[Exception]] = (),
        extend_child_of: bool = True,
    ):
        if extend_child_of:
            child_of = [TypeError, ValueError, *child_of]

        def _safe_cast(value: U) -> T | None:
            return safe_cast(self._caster, value, *child_of, default=None)

        return Caster(
            _safe_cast,
            name=f"{self._name}.safe_cast",
        )

    def or_[S](
        self, caster: Callable[[U], S], *childof: type[Exception]
    ) -> Caster[U, T | S]:
        """Returns a new caster that tries the original caster first, then the provided caster if the first fails."""

        if not childof:
            childof = (TypeError, ValueError)

        def _or(value: U) -> T | S:
            try:
                return self._caster(value)
            except childof:
                return caster(value)

        return Caster(
            _or,
            name=f"{self._name}.or_({getattr(caster, '__name__', type(caster).__name__)})",
        )

    @staticmethod
    def isinstance_or_cast[Arg](
        expects: type[T], onmiss: Callable[[Arg], U]
    ) -> Caster[Arg | T, T | U]:
        wrapped = isinstance_or_cast(expects, onmiss)
        return Caster(wrapped)

    def with_rule(
        self, rule: Callable[[T], bool], rulename: str | None = None
    ) -> Caster[U, T]:
        """Returns a new caster that applies the given rule to the result."""

        if rulename is None:
            if isinstance(rule, LambdaType) and rule.__name__ == "<lambda>":
                raise ValueError(
                    "Rule name must be provided for lambda functions"
                )

            rulename = rule.__name__

        def _with_rule(value: U) -> T:
            result = self._caster(value)
            if not rule(result):
                raise InvalidCast(
                    sentence(
                        f"result {result} does not satisfy the rule {squote(rulename)}"
                    ),
                    result,
                )
            return result

        return Caster(_with_rule, name=f"{self._name}.with_rule({rulename})")


def _extract_return_type(
    wrapper: Callable[[Any], Any],
    return_type: type[Any] | None,
) -> type[Any] | None:
    if return_type is not None:
        return return_type
    elif isinstance(wrapper, MethodType):
        return getattr(wrapper.__func__, "__annotations__", {}).get(
            "return", None
        )
    elif isinstance(wrapper, FunctionType):
        return getattr(wrapper, "__annotations__", {}).get("return", None)
    elif callable(wrapper) and not isinstance(wrapper, type):
        annotated = wrapper.__call__
        return getattr(annotated, "__annotations__", {}).get("return", None)
    elif isinstance(wrapper, type):  # pyright: ignore[reportUnnecessaryIsInstance]
        return wrapper
    else:
        return Any  # pyright: ignore[reportUnreachable]


def wrap_result_with[**P, T, U](
    wrapper: Callable[[T], U],
    return_type: type[U] | None = None,
) -> Callable[[Callable[P, T]], Callable[P, U]]:
    """
    Decorator factory that wraps the result of a function with a specified wrapper.
    This decorator modifies the return value of the decorated function by passing it
    through the provided `wrapper` callable. Optionally, it can update the function's
    return type annotation.
    Args:
        wrapper (Callable[[T], U]): A callable that takes the original return value
            and returns the wrapped value.
        return_type (type[U] | None, optional): The type to set as the return annotation
            for the decorated function. If None, attempts to infer from the wrapper.
    Returns:
        Callable[[Callable[P, T]], Callable[P, U]]: A decorator that wraps the result
        of the target function with `wrapper`.
    Example:
        >>> @wrap_result_with(str)
        ... def get_number() -> int:
        ...     return 42
        >>> get_number()
        '42'
    """

    def _wrap(func: Callable[P, T]) -> Callable[P, U]:
        func.__annotations__["return"] = _extract_return_type(
            wrapper, return_type
        )

        @functools.wraps(func)
        def __wrap(*args: P.args, **kwargs: P.kwargs) -> U:
            result = func(*args, **kwargs)
            return wrapper(result)

        return __wrap

    return _wrap


def awrap_result_with[**P, T, U](
    wrapper: Callable[[T], Coroutine[Any, Any, U]],
    return_type: type[U] | None = None,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, T]]], Callable[P, Coroutine[Any, Any, U]]
]:
    """
    Decorator factory that wraps the result of an asynchronous function with a specified wrapper.
    This decorator modifies the return value of the decorated asynchronous function by passing it
    through the provided `wrapper` callable, which must return an awaitable. Optionally, it can update the function's
    return type annotation.
    Args:
        wrapper (Callable[[T], Awaitable[U]]): A callable that takes the original return value
            and returns an awaitable that resolves to the wrapped value.
        return_type (type[U] | None, optional): The type to set as the return annotation
            for the decorated function. If None, attempts to infer from the wrapper.
    Returns:
        Callable[[Callable[P, T]], Callable[P, Coroutine[Any, Any, U]]]:
            A decorator that wraps the result of the target asynchronous function with `wrapper`.
    Example:
        >>> @awrap_result_with(lambda x: asyncio.sleep(0.1, result=str(x)))
        ... async def get_number() -> int:
        ...     return 42
        >>> await get_number()
        '42'
    """

    def _wrap(
        func: Callable[P, Coroutine[Any, Any, T]],
    ) -> Callable[P, Coroutine[Any, Any, U]]:
        func.__annotations__["return"] = _extract_return_type(
            wrapper, return_type
        )

        @functools.wraps(func)
        async def __wrap(*args: P.args, **kwargs: P.kwargs) -> U:
            result = await func(*args, **kwargs)
            return await wrapper(result)

        return __wrap

    return _wrap



from typing import Self

from escudeiro.escudeiro_pyrs import url
from escudeiro.url.mixins import Wrapped


class Netloc(Wrapped[url.Netloc]):
    def __init__(self, netloc: str) -> None:
        super().__init__(url.Netloc(netloc))

    @property
    def username(self) -> str | None:
        return self.internal.username

    @property
    def password(self) -> str | None:
        return self.internal.password

    @property
    def port(self) -> int | None:
        return self.internal.port

    @port.setter
    def port(self, port: int) -> None:
        self.internal.port = port

    @property
    def host(self) -> str:
        return self.internal.host

    def encode(self) -> str:
        return self.internal.encode()

    def parse(self, netloc: str) -> Self:
        self.internal.parse(netloc)
        return self

    def set(
        self,
        host: str | None = None,
        port: int | None = None,
        username: str | None = None,
        password: str | None = None,
    ) -> Self:
        self.internal.set(host, port, username, password)
        return self

    def merge(self, other: Self) -> Self:
        new_instance = object.__new__(type(self))
        new_instance.internal = self.internal.merge(other.internal)
        return new_instance

    def merge_left(self, other: Self) -> Self:
        new_instance = object.__new__(type(self))
        new_instance.internal = self.internal.merge_left(other.internal)
        return new_instance

    def merge_inplace(self, other: Self) -> Self:
        self.internal.merge_inplace(other.internal)
        return self

    @classmethod
    def from_args(
        cls,
        host: str,
        username: str | None = None,
        password: str | None = None,
        port: int | None = None,
    ) -> Self:
        self = object.__new__(cls)
        self.internal = url.Netloc.from_args(
            host=host, port=port, username=username, password=password
        )
        return self

    def copy(self) -> Self:
        new_instance = object.__new__(type(self))
        new_instance.internal = self.internal.copy()
        return new_instance



import re
from collections.abc import Callable, Sequence
from typing import Any, ForwardRef, NoReturn, TypeVar, get_args, get_origin

from .typedef import UNINITIALIZED, DisassembledType, TypeNode


def disassemble_type(annotation: Any) -> DisassembledType:
    type_ = (
        annotation
        if not isinstance(annotation, str)
        else ForwardRef(annotation)
    )

    node = make_node(type_)
    _, type_vars = extract_typevar(node)
    return DisassembledType(
        type_,
        get_origin(type_),
        get_args(type_),
        type_vars,
        node,
    )


def make_node(annotation: type | ForwardRef) -> TypeNode:
    root_node = TypeNode(get_origin(annotation) or annotation)
    stack: list[tuple[TypeNode, Any]] = [(root_node, annotation)]

    while stack:
        current_node, current_annotation = stack.pop()
        if get_origin(current_annotation) is None:
            continue

        type_args = get_args(current_annotation)
        for arg in type_args:
            if _arg := get_origin(arg):
                orig = _arg
            elif isinstance(arg, str):
                orig = ForwardRef(arg)
            else:
                orig = arg
            arg_node = TypeNode(orig)
            current_node.args.append(arg_node)
            stack.append((arg_node, arg))

    return root_node


def extract_typevar(
    node: TypeNode,
) -> tuple[TypeNode, Sequence[TypeNode]]:
    """Receives a type and returns a tuple with the original type and what typevars are part of it and at what depth"""
    if isinstance(node.type_, ForwardRef):
        return node, ()

    stack = [node]
    type_vars: list[TypeNode] = []
    while stack:
        current = stack.pop()
        if isinstance(current.type_, TypeVar):
            type_vars.append(current)
        if current.args:
            for child in current.args:
                stack.append(child)

    return node, type_vars


def get_forwardrefs(
    node: TypeNode,
) -> tuple[TypeNode, Sequence[TypeNode]]:
    stack = [node]
    type_vars = []
    depth = 0
    while stack:
        current = stack.pop()
        if isinstance(current.type_, ForwardRef):
            type_vars.append(current)
        if current.args:
            depth += 1
            for child in current.args:
                stack.append(child)

    return node, type_vars


def rebuild_type(origin: type, args: Sequence[Any]) -> type:
    if not hasattr(origin, "__getitem__"):
        raise TypeError("Unable to support rebuild, type has no __getitem__")
    try:
        return origin.__getitem__(*args)
    except TypeError:
        return origin[args]  # pyright: ignore[reportIndexIssue]


def rebuild_type_from_depth(node: TypeNode) -> type:
    stack = [(node, None)]
    rebuilt_types = {}

    while stack:
        current_node, rebuilt_type = stack.pop()

        if isinstance(current_node.type_, ForwardRef) or not current_node.args:
            # If the current node is a ForwardRef, use it as the rebuilt type.
            # If the current node has no child nodes, it's a basic type.
            rebuilt_types[current_node] = current_node.type_
        elif all(child in rebuilt_types for child in current_node.args):
            # If all child nodes have been rebuilt, reconstruct the type.
            arg_types = [rebuilt_types[child] for child in current_node.args]
            rebuilt_types[current_node] = rebuild_type(
                current_node.type_, tuple(arg_types)
            )
        else:
            # Push the current node back onto the stack and push its children.
            stack.append((current_node, rebuilt_type))
            for child in current_node.args:
                if child not in rebuilt_types:
                    stack.append((child, None))

    return rebuilt_types[node]


def frozen_setattr(self: object, name: str, value: Any):
    if getattr(self, name, UNINITIALIZED) is UNINITIALIZED:
        return object.__setattr__(self, name, value)
    raise AttributeError(
        f"Class {type(self)} is frozen, and attribute {name} cannot be set"
    )


def frozen_delattr(self: object, name: str) -> NoReturn:
    raise AttributeError(
        f"Class {type(self)} is frozen, and attribute {name} cannot be deleted"
    )


def frozen[T](cls: type[T]) -> type[T]:
    cls.__setattr__ = frozen_setattr
    cls.__delattr__ = frozen_delattr
    return cls


def indent(string: str, *, skip_line: bool = False) -> str:
    returnstr = f"    {string}"
    if skip_line:
        returnstr = "\n" + returnstr
    return returnstr


_sentinel = object()


def stamp_func(item: Callable | classmethod | staticmethod):
    to_stamp = item
    if isinstance(item, classmethod | staticmethod):
        to_stamp = item.__func__
    object.__setattr__(to_stamp, "__squire_func__", True)


def implements(cls: type, name: str):
    attr = getattr(cls, name, _sentinel)
    if attr is _sentinel:
        return False

    if hasattr(attr, "__squire_func__"):
        return False

    if func := getattr(attr, "__func__", None):
        if hasattr(func, "__squire_func__"):
            return False
    return all(
        getattr(base_cls, name, None) is not attr for base_cls in cls.mro()[1:]
    )


_dunder_regex = re.compile("__[a-zA-Z][a-zA-Z0-9]*__")


def is_dunder(string: str) -> bool:
    return _dunder_regex.match(string) is not None


def remove_left_underscores(string: str) -> str:
    if is_dunder(string):
        return string
    return string.lstrip("_")



from collections import OrderedDict
from collections.abc import Sequence
from enum import Enum
from typing import Any, override

from escudeiro.misc.iterx import next_or
from escudeiro.misc.lazy import lazymethod
from escudeiro.misc.strings import to_camel, to_kebab, to_pascal, to_snake


class StrEnum(str, Enum):
    """A base class for string-based enums with enhanced functionality for aliases.

    This class provides a custom `_missing_` method to search for enum members based on aliases
    and includes a method for generating various string aliases for enum values (camel, pascal, kebab, etc.)
    """

    @classmethod
    @override
    def _missing_(cls, value: object) -> Any:
        return next_or(item for item in cls if value in item.get_aliases())

    @lazymethod
    def get_aliases(self) -> Sequence[str]:
        return tuple(
            OrderedDict.fromkeys(
                (
                    self.value,
                    self.name,
                    self.name.upper(),
                    self.name.lower(),
                    to_camel(self.name),
                    to_pascal(self.name),
                    to_kebab(self.name),
                )
            )
        )


class ValueEnum(StrEnum):
    """A subclass of `StrEnum` where the string representation is based on the enum's value.

    This class overrides the `__str__` method to return the value of the enum member instead of its name.
    """

    @override
    def __str__(self) -> str:
        return self.value


class NameEnum(StrEnum):
    """A subclass of `StrEnum` where the string representation is based on the enum's name.

    This class overrides the `__str__` method to return the name of the enum member instead of its value.
    """

    @override
    def __str__(self) -> str:
        return self.name


class SnakeEnum(StrEnum):
    """A subclass of `StrEnum` that automatically generates enum values in snake_case.

    This class overrides the `_generate_next_value_` method to automatically generate a
    snake_case value based on the enum name.
    """

    @staticmethod
    @override
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[Any]
    ) -> Any:
        return to_snake(name)


class CamelEnum(StrEnum):
    """A subclass of `StrEnum` that automatically generates enum values in camelCase.

    This class overrides the `_generate_next_value_` method to automatically generate a
    camelCase value based on the enum name.
    """

    @staticmethod
    @override
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[Any]
    ) -> Any:
        return to_camel(name)


class PascalEnum(StrEnum):
    """A subclass of `StrEnum` that automatically generates enum values in PascalCase.

    This class overrides the `_generate_next_value_` method to automatically generate a
    PascalCase value based on the enum name.
    """

    @staticmethod
    @override
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[Any]
    ) -> Any:
        return to_pascal(name)


class KebabEnum(StrEnum):
    """A subclass of `StrEnum` that automatically generates enum values in kebab-case.

    This class overrides the `_generate_next_value_` method to automatically generate a
    kebab-case value based on the enum name.
    """

    @staticmethod
    @override
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[Any]
    ) -> Any:
        return to_kebab(name)


# If combining many of those utilities, inherit the NameEnum or ValueEnum first,
# to avoid overriding the automatic name generation with Enum's default one.



import os
from collections.abc import Iterator, MutableMapping
from typing import override

from escudeiro.data import data, field
from escudeiro.exc import KeyAlreadyRead


@data
class EnvMapping(MutableMapping[str, str]):
    mapping: MutableMapping[str, str] = os.environ
    already_read: set[str] = field(default_factory=set)

    @override
    def __getitem__(self, key: str, /) -> str:
        val = self.mapping[key]
        self.already_read.add(key)
        return val

    @override
    def __setitem__(self, key: str, value: str, /) -> None:
        if key in self.already_read:
            raise KeyAlreadyRead(f"{key} already read, cannot change its value")
        self.mapping[key] = value

    @override
    def __delitem__(self, key: str, /) -> None:
        if key in self.already_read:
            raise KeyAlreadyRead(f"{key} already read, cannot delete it")
        del self.mapping[key]

    @override
    def __iter__(self) -> Iterator[str]:
        yield from self.mapping

    @override
    def __len__(self) -> int:
        return len(self.mapping)


DEFAULT_MAPPING = EnvMapping()



import threading
from types import TracebackType
from typing import override

from escudeiro.context.context import AsyncContext, Context
from escudeiro.context.interfaces import AtomicAdapter, AtomicAsyncAdapter
from escudeiro.data import data, private
from escudeiro.data.helpers import call_init
from escudeiro.lazyfields import mark_class


@mark_class(threading.RLock)
@data(frozen=False)
class BoundContext[T](Context[T]):
    """A context manager for managing atomic transactions with an adapter."""

    adapter: AtomicAdapter[T]
    context: Context[T]

    _already_inited: bool = private(initial=False)

    def __new__(cls, adapter: AtomicAdapter[T], context: Context[T]):
        if not isinstance(context, cls) or adapter is not context.adapter:
            self = object.__new__(cls)
            return self
        return context

    def __init__(self, adapter: AtomicAdapter[T], context: Context[T]) -> None:
        """
        Initialize a BoundContext instance.

        Args:
            adapter (AtomicAdapter[T]): The atomic adapter for managing transactions.
            context (Context[T]): The underlying context to manage.
        """
        if hasattr(self, "_already_inited"):
            return
        call_init(self, adapter, context)
        self._already_inited = True

    @override
    def acquire(self) -> T:
        """
        Acquire the context for a transaction and begin an atomic operation.

        Returns:
            T: The acquired context.
        """
        with self._lock:
            client = self.context.acquire()
            self.stack += 1
            self.adapter.begin(client)
        return client

    @override
    def release(self, commit: bool = True):
        """
        Release the context, committing or rolling back the transaction if needed.

        Args:
            commit (bool, optional): Whether to commit the transaction. Defaults to True.
        """
        with self._lock:
            if self.stack == 1:
                if self.adapter.in_atomic(self.context.client):
                    if commit:
                        self.adapter.commit(self.context.client)
                    else:
                        self.adapter.rollback(self.context.client)
                self.context.release()
            self.stack -= 1

    @override
    def __exit__(
        self,
        *exc: tuple[
            type[BaseException] | None, Exception | None, TracebackType | None
        ],
    ):
        """
        Exit the context manager and release the context.

        Args:
            *exc: Exception information.
        """
        self.release(not any(exc))


@data(frozen=False)
class AsyncBoundContext[T](AsyncContext[T]):
    """An asynchronous context manager for managing atomic transactions with an async adapter."""

    adapter: AtomicAsyncAdapter[T]
    context: AsyncContext[T]
    _already_inited: bool = private(initial=False)

    def __new__(cls, adapter: AtomicAsyncAdapter[T], context: AsyncContext[T]):
        if not isinstance(context, cls) or adapter is not context.adapter:
            self = object.__new__(cls)
            return self
        return context

    def __init__(
        self, adapter: AtomicAsyncAdapter[T], context: AsyncContext[T]
    ) -> None:
        """
        Initialize an AsyncBoundContext instance.

        Args:
            adapter (AtomicAsyncAdapter[T]): The async atomic adapter for managing transactions.
            context (AsyncContext[T]): The underlying async context to manage.
        """
        if hasattr(self, "_already_inited"):
            return
        call_init(self, adapter, context)
        self._already_inited = True

    @override
    async def acquire(self) -> T:
        """
        Acquire the async context for a transaction and begin an atomic operation.

        Returns:
            T: The acquired async context.
        """
        async with self.lock:
            client = await self.context.acquire()
            if self.stack == 0:
                await self.adapter.begin(client)
            self.stack += 1
            return client

    @override
    async def release(self, commit: bool = True):
        """
        Release the async context, committing or rolling back the transaction if needed.

        Args:
            commit (bool, optional): Whether to commit the transaction. Defaults to True.
        """
        async with self.lock:
            client = await self.context.client()
            if self.stack == 1:
                if commit:
                    await self.adapter.commit(client)
                else:
                    await self.adapter.rollback(client)
            self.stack -= 1
        await self.context.release()

    @override
    async def __aexit__(
        self,
        *exc: tuple[
            type[BaseException] | None,
            BaseException | None,
            TracebackType | None,
        ],
    ):
        """
        Exit the async context manager and release the async context.

        Args:
            *exc: Exception information.
        """
        await self.release(not any(exc))



from typing import Any, TypeVar

from escudeiro.config.interface import ConfigLike
from escudeiro.data import data
from escudeiro.lazyfields import lazyfield

from .factory import DEFAULT_CONFIG, AdapterConfigFactory

T = TypeVar("T")


@data
class CachedFactory:
    config: ConfigLike = DEFAULT_CONFIG

    @lazyfield
    def _cache(self) -> dict[tuple[str, type], Any]:
        return {}

    @lazyfield
    def _primary(self) -> dict[type, Any]:
        return {}

    @lazyfield
    def _internal(self) -> AdapterConfigFactory:
        return AdapterConfigFactory(self.config)

    def load(
        self,
        model_cls: type[T],
        __prefix__: str = "",
        __sep__: str = "__",
        __primary__: bool = False,
        *,
        presets: dict[str, Any] | None = None,
        **defaults: Any,
    ) -> T:
        if memo := self._cache.get((__prefix__, model_cls)):
            return memo
        if (memo := self._primary.get(model_cls)) and not __prefix__:
            return memo
        memo = self._internal.load(
            model_cls,
            __prefix__,
            __sep__,
            presets=presets,
            **defaults,
        )
        self._cache[(__prefix__, model_cls)] = memo
        if __primary__:
            self._primary[model_cls] = memo
        return memo



# pyright: reportUnusedParameter=false,reportMissingParameterType=false
from typing import Generic, TypeVar, override

import pytest

from escudeiro.misc.lazy import MethodType, lazymethod


def test_method_type_detection():
    """Test the internal _determine_method_type logic."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def self_only(self):
            self.counter += 1
            return f"self_only_{self.counter}"

        @lazymethod
        def hashable_args(self, a: str, b: int):
            self.counter += 1
            return f"hashable_{self.counter}"

        @lazymethod
        def unhashable_args(self, a: list):
            self.counter += 1
            return f"unhashable_{self.counter}"

        @lazymethod
        def unannotated_args(self, a):
            self.counter += 1
            return f"unannotated_{self.counter}"

    # Extract lazymethod instances
    self_only_method = Test.__dict__["self_only"]
    hashable_method = Test.__dict__["hashable_args"]
    unhashable_method = Test.__dict__["unhashable_args"]
    unannotated_method = Test.__dict__["unannotated_args"]

    # Check method types
    assert self_only_method._method_type == MethodType.SELF_ONLY
    assert hashable_method._method_type == MethodType.HASHABLE_ARGS
    assert unhashable_method._method_type == MethodType.UNKNOWN_OR_UNHASHABLE
    assert unannotated_method._method_type == MethodType.UNKNOWN_OR_UNHASHABLE


def test_container_selection():
    """Test container type selection based on method type."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def self_only(self):
            self.counter += 1
            return f"self_only_{self.counter}"

        @lazymethod
        def hashable_args(self, a: str, b: int):
            self.counter += 1
            return f"hashable_{self.counter}"

        @lazymethod
        def unhashable_args(self, a: list):
            self.counter += 1
            return f"unhashable_{self.counter}"

    # Extract lazymethod instances
    self_only_method = Test.__dict__["self_only"]
    hashable_method = Test.__dict__["hashable_args"]
    unhashable_method = Test.__dict__["unhashable_args"]

    # Check container types
    assert self_only_method._container is None
    assert hashable_method._container is dict
    assert unhashable_method._container is list


def test_self_only_caching():
    """Test caching with no parameters (only self is used as cache key)."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate(self):
            self.counter += 1
            return f"result_{self.counter}"

    # Test with a single instance
    t1 = Test()
    assert t1.calculate() == "result_1"
    assert t1.calculate() == "result_1"  # Cached result
    assert t1.counter == 1  # Function called only once

    # Verify the private attribute exists and holds the value
    private_name = lazymethod.get_private("calculate")
    assert hasattr(t1, private_name)
    assert getattr(t1, private_name) == "result_1"

    # Test with a different instance
    t2 = Test()
    assert t2.calculate() == "result_1"  # New instance, new calculation
    assert t2.counter == 1

    # Verify initialization status
    assert lazymethod.is_initialized(t1, "calculate")
    assert lazymethod.is_initialized(t2, "calculate")
    t3 = Test()
    assert not lazymethod.is_initialized(t3, "calculate")


def test_hashable_args_caching():
    """Test with annotated hashable parameters that are used as cache keys."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate_with_params(self, a: str, b: int, c: bool = True):
            self.counter += 1
            return f"{a}_{b}_{int(c)}_{self.counter}"

    t = Test()

    # Basic parameter caching
    assert t.calculate_with_params("test", 1) == "test_1_1_1"
    assert t.calculate_with_params("test", 1) == "test_1_1_1"  # Cached
    assert (
        t.calculate_with_params("other", 1) == "other_1_1_2"
    )  # Different first param
    assert (
        t.calculate_with_params("test", 2) == "test_2_1_3"
    )  # Different second param
    assert t.counter == 3

    # Test with default and explicit kwargs
    assert (
        t.calculate_with_params("test", 1, False) == "test_1_0_4"
    )  # Different default
    assert (
        t.calculate_with_params("test", 1, c=False) == "test_1_0_4"
    )  # Should be cached
    assert (
        t.calculate_with_params("test", b=1, c=True) == "test_1_1_1"
    )  # Should match first call
    assert t.counter == 4

    # Verify the private attribute exists and holds a dict with proper keys
    private_name = lazymethod.get_private("calculate_with_params")
    assert hasattr(t, private_name)
    cache_dict = getattr(t, private_name)
    assert isinstance(cache_dict, dict)

    # Check specific cache keys
    # The key format is (args, tuple(kwargs.items()))
    key1 = (("a", "test"), ("b", 1), ("c", True))  # a="test", b=1, no kwargs
    key2 = (("a", "test"), ("b", 1), ("c", False))  # a="test", b=1, c=False

    assert cache_dict[frozenset(key1)] == "test_1_1_1"
    assert cache_dict[frozenset(key2)] == "test_1_0_4"


def test_unhashable_args_caching():
    """Test with unhashable or unannotated parameters."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate_with_list(self, items: list):
            self.counter += 1
            return f"list_{len(items)}_{self.counter}"

    t = Test()

    # List parameter (unhashable)
    list1 = [1, 2, 3]
    list2 = [1, 2, 3]  # Same content but different object

    assert t.calculate_with_list(list1) == "list_3_1"
    assert t.calculate_with_list(list1) == "list_3_1"  # Same object, cached
    assert (
        t.calculate_with_list(list2) == "list_3_1"
    )  # Different object, but equality exists
    assert t.counter == 1

    # Verify the private attribute exists and holds a list with proper items
    private_name = lazymethod.get_private("calculate_with_list")
    assert hasattr(t, private_name)
    cache_list = getattr(t, private_name)
    assert isinstance(cache_list, list)
    assert len(cache_list) == 1

    # Cache items are tuples of ((args, kwargs_tuple), result)
    assert cache_list[0][0] == (("items", list1),)
    assert cache_list[0][1] == "list_3_1"


def test_mixed_parameter_types():
    """Test methods with a mix of hashable and unhashable parameters."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate_mixed(self, name: str, items: list):
            self.counter += 1
            return f"{name}_{len(items)}_{self.counter}"

    t = Test()
    list1 = [1, 2, 3]
    list2 = [1, 2, 3]  # Different object, same content

    assert t.calculate_mixed("test", list1) == "test_3_1"
    assert (
        t.calculate_mixed("test", list1) == "test_3_1"
    )  # Same objects, cached
    assert (
        t.calculate_mixed("test", list2) == "test_3_1"
    )  # Different list object, equality still exists
    assert t.calculate_mixed("other", list1) == "other_3_2"  # Different string
    assert t.counter == 2

    # Since there's an unhashable parameter, container should be a list
    private_name = lazymethod.get_private("calculate_mixed")
    cache_list = getattr(t, private_name)
    assert isinstance(cache_list, list)
    assert len(cache_list) == 2


def test_complex_hashable_types():
    """Test caching with complex hashable type annotations."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate_complex(self, a: tuple[str, int], b: frozenset):
            self.counter += 1
            return f"complex_{self.counter}"

    t = Test()

    tup = ("test", 1)
    fs = frozenset([1, 2, 3])

    assert t.calculate_complex(tup, fs) == "complex_1"
    assert t.calculate_complex(tup, fs) == "complex_1"  # Cached
    assert (
        t.calculate_complex(("test", 2), fs) == "complex_2"
    )  # Different tuple
    assert t.counter == 2

    # Since parameters are hashable, container should be a dict
    private_name = lazymethod.get_private("calculate_complex")
    cache = getattr(t, private_name)
    assert isinstance(cache, dict)


def test_nested_container_types():
    """Test with nested container types to verify recursive hashability check."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def with_nested_tuple_dict(self, a: tuple[dict[str, int], ...]):
            self.counter += 1
            return f"nested_tuple_dict_{self.counter}"

        @lazymethod
        def with_nested_list_dict(self, a: list[dict[str, int]]):
            self.counter += 1
            return f"nested_list_dict_{self.counter}"

    t = Test()

    # Tuple containing dicts (hashable overall)
    tup_dict = ({"a": 1}, {"b": 2})
    assert t.with_nested_tuple_dict(tup_dict) == "nested_tuple_dict_1"
    assert t.with_nested_tuple_dict(tup_dict) == "nested_tuple_dict_1"  # Cached

    # Although parameters' origins are hashable, args are not, container should be list
    private_name = lazymethod.get_private("with_nested_tuple_dict")
    cache = getattr(t, private_name)
    assert isinstance(cache, list)

    # List containing dicts (unhashable)
    list_dict = [{"a": 1}, {"b": 2}]
    assert t.with_nested_list_dict(list_dict) == "nested_list_dict_2"
    assert (
        t.with_nested_list_dict(list_dict) == "nested_list_dict_2"
    )  # Cached by identity

    # Since parameters are unhashable, container should be a list
    private_name = lazymethod.get_private("with_nested_list_dict")
    cache = getattr(t, private_name)
    assert isinstance(cache, list)


def test_kwargs_handling():
    """Test handling of keyword arguments in caching."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def with_kwargs(self, a: str, **kwargs: int):
            self.counter += 1
            return f"kwargs_{self.counter}"

    t = Test()

    assert t.with_kwargs("test") == "kwargs_1"
    assert t.with_kwargs("test") == "kwargs_1"  # Cached
    assert t.with_kwargs("test", x=1) == "kwargs_2"  # Different kwargs
    assert t.with_kwargs("test", x=1) == "kwargs_2"  # Cached
    assert t.with_kwargs("test", x=1, y=2) == "kwargs_3"  # More kwargs
    assert (
        t.with_kwargs("test", y=2, x=1) == "kwargs_3"
    )  # Same kwargs, different order
    assert t.counter == 3


def test_descriptor_behavior():
    """Test the descriptor behavior of lazymethod."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate(self):
            self.counter += 1
            return f"result_{self.counter}"

    # Access via class should return the descriptor itself
    descriptor = Test.calculate
    assert isinstance(descriptor, lazymethod)

    # Access via instance should return a callable
    t = Test()
    method = t.calculate
    assert callable(method)

    # Calling the method should work
    assert method() == "result_1"


def test_internal_methods():
    """Test internal methods of lazymethod."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate(self, a: str):
            self.counter += 1
            return f"{a}_{self.counter}"

    t = Test()

    # Get the descriptor and method name
    descriptor = Test.__dict__["calculate"]
    private_name = descriptor.private_name

    # Test _get when no value exists
    result = descriptor._get(t, private_name, ("test",), {})
    assert result is None

    # Test _set
    result = descriptor._set(t, "test")
    assert result == "test_1"

    # Test _get after value exists
    result = descriptor._get(t, private_name, ("test",), {})
    assert result == "test_1"


def test_inheritance():
    """Test lazymethod behavior with class inheritance."""

    class Base:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate(self, a: str):
            self.counter += 1
            return f"base_{a}_{self.counter}"

    class Derived(Base):
        @lazymethod
        def calculate(self, a: str):
            self.counter += 1
            return f"derived_{a}_{self.counter}"

    b = Base()
    d = Derived()

    assert b.calculate("test") == "base_test_1"
    assert d.calculate("test") == "derived_test_1"

    # Each instance should have its own cache
    assert lazymethod.is_initialized(b, "calculate")
    assert lazymethod.is_initialized(d, "calculate")


def test_reset_functionality():
    """Test resetting cached values."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def self_only(self):
            self.counter += 1
            return f"self_only_{self.counter}"

        @lazymethod
        def with_args(self, a: str):
            self.counter += 1
            return f"{a}_{self.counter}"

        @lazymethod
        def with_unhashable(self, items: list):
            self.counter += 1
            return f"list_{self.counter}"

    t = Test()

    # Initial calculations
    assert t.self_only() == "self_only_1"
    assert t.with_args("test") == "test_2"
    list1 = [1, 2, 3]
    assert t.with_unhashable(list1) == "list_3"

    # Test manual reset by deleting private attributes
    private_name1 = lazymethod.get_private("self_only")
    delattr(t, private_name1)
    assert not lazymethod.is_initialized(t, "self_only")
    assert t.self_only() == "self_only_4"  # Recalculated

    # Reset specific parameter
    private_name2 = lazymethod.get_private("with_args")
    cache_dict = getattr(t, private_name2)
    key = (("a", "test"),)
    del cache_dict[frozenset(key)]
    assert t.with_args("test") == "test_5"  # Recalculated

    # Reset unhashable parameter
    private_name3 = lazymethod.get_private("with_unhashable")
    cache_list = getattr(t, private_name3)
    cache_list.clear()
    assert t.with_unhashable(list1) == "list_6"  # Recalculated


def test_is_hashable_function():
    """Test the _is_hashable function's core logic."""

    class CustomClass:
        pass

    class HashableClass:
        @override
        def __hash__(self):
            return 42

    # Create a lazymethod instance to access _is_hashable
    class TestClass:
        @lazymethod
        def dummy(self):
            pass

    descriptor = TestClass.__dict__["dummy"]

    # Test various types
    assert descriptor.is_hashable(str) is True
    assert descriptor.is_hashable(int) is True
    assert descriptor.is_hashable(tuple) is True
    assert descriptor.is_hashable(list) is False
    assert descriptor.is_hashable(dict[str, int]) is False
    assert descriptor.is_hashable(tuple[int, str]) is True
    assert descriptor.is_hashable(list[int]) is False
    assert descriptor.is_hashable(HashableClass) is True
    assert descriptor.is_hashable(CustomClass) is True  # Classes are hashable

    # Test recursive types
    assert descriptor.is_hashable(tuple[dict[str, int], ...]) is False
    assert descriptor.is_hashable(tuple[tuple[int, str], ...]) is True


def test_generic_types():
    """Test with generic type annotations."""
    T = TypeVar("T")

    class GenericTest(Generic[T]):
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate(self, a: T):
            self.counter += 1
            return f"generic_{self.counter}"

    t_int = GenericTest[int]()
    t_str = GenericTest[str]()

    assert t_int.calculate(1) == "generic_1"
    assert t_int.calculate(1) == "generic_1"  # Cached
    assert t_int.calculate(2) == "generic_2"  # Different value

    assert t_str.calculate("a") == "generic_1"
    assert t_str.calculate("a") == "generic_1"  # Cached
    assert t_str.calculate("b") == "generic_2"  # Different value


def test_descriptor_name_setting():
    """Test that __set_name__ properly sets public and private names."""

    class Test:
        def __init__(self) -> None:
            pass

        @lazymethod
        def original_name(self):
            return "value"

    # Access the descriptor directly
    descriptor = Test.__dict__["original_name"]
    assert descriptor.public_name == "original_name"
    assert descriptor.private_name == "_lazymethod_original_name_"

    # Test format
    assert lazymethod.format_ == "_lazymethod_{method_name}_"
    assert lazymethod.get_private("test_method") == "_lazymethod_test_method_"


def test_overloaded_method():
    """Test lazymethod with type-overloaded methods."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def calculate(self, param: str | int):
            self.counter += 1
            if isinstance(param, str):
                return f"str_{param}_{self.counter}"
            else:
                return f"int_{param}_{self.counter}"

    t = Test()

    assert t.calculate("test") == "str_test_1"
    assert t.calculate("test") == "str_test_1"  # Cached
    assert t.calculate(123) == "int_123_2"
    assert t.calculate(123) == "int_123_2"  # Cached
    assert t.counter == 2


def test_edge_cases():
    """Test edge cases and potential issues."""

    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def empty_tuple(self, a: tuple = ()):
            self.counter += 1
            return f"empty_{self.counter}"

        @lazymethod
        def none_value(self, a: str | None = None):
            self.counter += 1
            return f"none_{self.counter}"

    t = Test()

    # Empty tuples and None values should work as expected
    assert t.empty_tuple() == "empty_1"
    assert t.empty_tuple() == "empty_1"  # Cached
    assert t.empty_tuple(()) == "empty_1"  # Same as default

    assert t.none_value() == "none_2"
    assert t.none_value() == "none_2"  # Cached
    assert t.none_value(None) == "none_2"  # Same as default


def test_exception_does_not_corrupt_cache():
    class Test:
        def __init__(self) -> None:
            self.counter = 0

        @lazymethod
        def self_only(self):
            # Container is None for SELF_ONLY methods.
            self.counter += 1
            if self.counter <= 1:
                raise ValueError("Exception in self_only")
            return self.counter

        @lazymethod
        def hashable(self, a: int, b: str) -> str:
            # Container is dict because args are hashable.
            self.counter += 1
            if self.counter == 2:
                raise ValueError("Exception in hashable")
            return f"{a}_{b}_{self.counter}"

        @lazymethod
        def mixed(self, a: int, b: list[str], c, d=True) -> str:
            # Container is list because of unhashable argument in b.
            self.counter += 1
            if self.counter == 2:
                raise ValueError("Exception in mixed")
            return f"{a}_{b}_{c}_{d}"

    obj = Test()

    # --- Test SELF_ONLY method ---
    # First call should raise, and since container is None the
    # private attribute shouldn't be created.
    private_self = lazymethod.get_private("self_only")
    with pytest.raises(ValueError):
        _ = obj.self_only()

    assert not hasattr(obj, private_self), (
        "Cache attribute exists even after exception for self_only"
    )

    # Subsequent call succeeds and caches the result.
    result_self = obj.self_only()  # Now counter becomes 2 and no exception
    assert result_self == 2
    assert hasattr(obj, private_self), (
        "Cache attribute not created after successful self_only call"
    )
    # Further calls return the same cached value.
    assert obj.self_only() == 2, (
        "Cached self_only value did not match expected value"
    )

    # --- Test HASHABLE method ---
    # For hashable, container is a dict.
    private_hash = lazymethod.get_private("hashable")
    # First call with (1, "a") should raise.
    obj.counter = 1  # reset counter to force error
    with pytest.raises(ValueError):
        _ = obj.hashable(1, "a")

    # No caching should have occurred for that argument combo.
    assert not hasattr(obj, private_hash), (
        "Cache attribute exists for hashable after exception"
    )

    # Call again; now the call succeeds.
    # After self_only, counter == 2. The first hashable call increments counter from 2 to 3 and raises.
    # A new call will increment counter from 3 to 4.
    result_hash = obj.hashable(1, "a")
    # Expect the method to return a string incorporating the current counter.
    # With counter at 4, we expect "1_a_4".
    assert result_hash == "1_a_3", (
        f"Unexpected result from hashable: {result_hash}"
    )
    # Now the caching container (a dict) should be present
    cache_dict = getattr(obj, private_hash)
    # The key is built via _signature_to_kwargs; for a two-parameter method,
    # the mapping will be {"a": 1, "b": "a"} and the key is frozenset of its items.
    key_hash = frozenset([("a", 1), ("b", "a")])
    assert cache_dict.get(key_hash) == "1_a_3", (
        "Cached value not found in hashable container"
    )

    # --- Test MIXED method ---
    # For mixed, container is a list (because of the unhashable list parameter).
    private_mixed = lazymethod.get_private("mixed")
    # First call with arguments (10, ["x"], "y") should raise.
    obj.counter = 1  # reset counter to trigger error

    with pytest.raises(ValueError):
        _ = obj.mixed(10, ["x"], "y")

    assert not hasattr(obj, private_mixed), (
        "Cache attribute exists for mixed after exception"
    )

    # Call again; the call should succeed.
    result_mixed = obj.mixed(10, ["x"], "y")
    # After hashable, counter should be 4 and mixed call increments it to 5.
    assert result_mixed == "10_['x']_y_True", (
        f"Unexpected result from mixed: {result_mixed}"
    )

    # Now the container (a list) should hold one entry.
    cache_list = getattr(obj, private_mixed)
    assert isinstance(cache_list, list)
    assert len(cache_list) == 1
    # For mixed, the key is built from the normalized signature.
    # Expected normalized mapping is {"a": 10, "b": ["x"], "c": "y", "d": True}.
    # The key in a list container is stored as a tuple of items (order determined by sorting of keys).
    expected_key = (("a", 10), ("b", ["x"]), ("c", "y"), ("d", True))
    cached_key, cached_value = cache_list[0]
    assert cached_key == expected_key, (
        f"Unexpected key in mixed cache: {cached_key}"
    )
    assert cached_value == "10_['x']_y_True", (
        "Cached value in mixed container does not match"
    )

    # --- Verify that further calls with the same arguments return the cached results ---
    obj.counter = 0
    assert obj.self_only() == 2, "Subsequent self_only call used cached value"
    assert obj.hashable(1, "a") == "1_a_3", (
        "Subsequent hashable call used cached value"
    )
    assert obj.mixed(10, ["x"], "y") == "10_['x']_y_True", (
        "Subsequent mixed call used cached value"
    )
    assert obj.counter == 0, "Counter changed on a cached call"



import asyncio

import pytest

from escudeiro.misc.monad import (
    AsyncCell,
    AsyncMonadHelper,
    BaseMonad,
    Either,
    LazyMonad,
    Left,
    Monad,
    Nothing,
    NullMonad,
    Right,
    cast_to_lazy,
    if_,
    make_lazy,
    resolve_lazy,
)


class TestBaseMonad:
    def test_get_value_and_repr(self):
        m = BaseMonad.pure(10)
        assert m.get_value() == 10
        assert repr(m) == "BaseMonad(10)"

    def test_pure_creates_instance(self):
        m = BaseMonad.pure("abc")
        assert isinstance(m, BaseMonad)
        assert m.get_value() == "abc"


class TestMonad:
    def test_map(self):
        m = Monad.pure(2)
        m2 = m.map(lambda x: x + 3)
        assert isinstance(m2, Monad)
        assert m2.get_value() == 5

    def test_bind(self):
        m = Monad.pure(2)

        def f(x: int):
            return Monad.pure(x * 10)

        m2 = m.bind(f)
        assert isinstance(m2, Monad)
        assert m2.get_value() == 20

    def test_maybe_returns_monad(self):
        m = Monad.pure(2)
        m2 = m.maybe(lambda x: x * 2)
        assert isinstance(m2, Monad)
        assert m2.get_value() == 4

    def test_maybe_returns_nothing(self):
        m = Monad.pure(2)
        m2 = m.maybe(lambda x: None)
        assert m2 is Nothing


class TestNullMonad:
    def test_singleton(self):
        n1 = NullMonad()
        n2 = NullMonad()
        assert n1 is n2
        assert n1.value is None

    def test_get_value_raises(self):
        n = NullMonad()
        with pytest.raises(ValueError, match="NullMonad has no value"):
            n.get_value()

    def test_bind_and_map_return_self(self):
        n = NullMonad()
        assert n.bind(lambda x: Monad.pure(1)) is n
        assert n.map(lambda x: 2) is n


class TestEither:
    def test_left_and_right(self):
        left = Left("err")
        r = Right(123)
        assert isinstance(left, Left)
        assert isinstance(r, Right)
        assert left.value == "err"
        assert r.value == 123

    def test_left_bind_and_map(self):
        left = Left("fail")
        assert left.bind(lambda x: Right(x + 1)) is left
        assert left.map(lambda x: x + 1) is left

    def test_right_bind_and_map(self):
        r = Right(10)
        r2 = r.bind(lambda x: Right(x * 2))
        assert isinstance(r2, Right)
        assert r2.value == 20
        r3 = r.map(lambda x: x + 5)
        assert isinstance(r3, Right)
        assert r3.value == 15

    def test_right_get_value(self):
        r = Right("abc")
        assert r.get_value() == "abc"

    def test_either_base_not_implemented(self):
        e = Either("lvalue")
        with pytest.raises(NotImplementedError):
            _ = e.bind(lambda x: x)
        with pytest.raises(NotImplementedError):
            _ = e.map(lambda x: x)


class TestAsyncCell:
    def test_async_cell_resolve(self):
        async def coro():
            await asyncio.sleep(0)
            return 42

        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(coro())
        cell = AsyncCell(future, lambda f: loop.run_until_complete(f))
        assert cell.resolve() == 42


class TestAsyncMonadHelper:
    def test_wrap_eager_and_lazy(self):
        async def coro(x: int):
            await asyncio.sleep(0)
            return x * 2

        loop = asyncio.get_event_loop()
        helper = AsyncMonadHelper(loop)
        eager = helper.wrap_eager(coro)
        lazy = helper.wrap_lazy(coro)
        m = eager(5)
        assert isinstance(m, Monad)
        assert m.get_value() == 10
        cell = lazy(6)
        assert isinstance(cell, AsyncCell)
        assert cell.resolve() == 12


class TestLazyMonad:
    def test_lazy_monad_map_and_bind(self):
        m = LazyMonad(2)
        m2 = m.map(lambda x: x + 3)
        assert isinstance(m2, LazyMonad)
        m3 = m2.bind(lambda x: Monad.pure(x * 2))
        assert isinstance(m3, LazyMonad)
        resolved = m3.resolve()
        assert isinstance(resolved, Monad)
        assert resolved.get_value() == 10

    def test_lazy_monad_resolve_without_binder(self):
        m = LazyMonad(7)
        result = m.resolve()
        assert isinstance(result, Monad)
        assert result.get_value() == 7

    def test_lazy_monad_badly_formed(self):
        parent = LazyMonad(1)
        m = LazyMonad(2, parent=parent, binder=None)
        with pytest.raises(TypeError, match="Badly formed LazyMonad"):
            _ = m.resolve()


class TestResolveLazy:
    def test_resolve_lazy(self):
        m = LazyMonad(3).map(lambda x: x + 1)
        result = resolve_lazy(m)
        assert isinstance(result, Monad)
        assert result.get_value() == 4


class TestCastToLazy:
    def test_cast_to_lazy(self):
        m = LazyMonad(1).map(lambda x: x + 1)
        lazy = cast_to_lazy(10, m)
        assert isinstance(lazy, LazyMonad)
        resolved = lazy.resolve()
        assert resolved.get_value() == 10


class TestMakeLazy:
    def test_make_lazy(self):
        def binder(x: int):
            return Monad.pure(x * 5)

        lazy_func = make_lazy(binder)
        m = lazy_func(2)
        assert isinstance(m, LazyMonad)
        result = m.resolve()
        assert result.get_value() == 10


class TestIf_:
    def test_if_true(self):
        def cond(x: int):
            return x > 0

        def ontrue(x: int):
            return x + 1

        onfalse = -1
        f = if_(cond, ontrue, onfalse)
        result = f(2)
        assert isinstance(result, Right)
        assert result.value == 3

    def test_if_false(self):
        def cond(x: int):
            return x < 0

        def ontrue(x: int):
            return x + 1

        onfalse = -1
        f = if_(cond, ontrue, onfalse)
        result = f(2)
        assert isinstance(result, Left)
        assert result.value == -1



import ast
import importlib
import os
import pathlib
import sys
import warnings
from collections import defaultdict
from collections.abc import Callable, Iterator, Sequence
from inspect import getsourcelines
from pathlib import Path
from types import ModuleType

from escudeiro.data import data
from escudeiro.lazyfields import lazyfield
from escudeiro.misc import filter_isinstance, lazymethod

type PathConverter = Callable[[Path], str]
type StrOrPath = str | Path


def make_modulename_converter(
    root: pathlib.Path, path_converter: PathConverter
) -> PathConverter:
    root_path = path_converter(root)

    def converter(path: Path) -> str:
        pathstr = path_converter(path)
        if root_path in pathstr:
            pathstr = pathstr.replace(root_path, root.name)
        return pathstr.removesuffix(".py").replace("/", ".")

    return converter


@data(frozen=False)
class AutoDiscoveryHelper:
    """
    Helper class for auto-discovery of modules or files within a directory.

    Attributes:
        root (pathlib.Path): Root directory to start discovery.
        look_on (Path | None): Optional specific directory to look on.
        exclude (Sequence[StrOrPath]): List of paths or filenames to exclude.
        exts (Sequence[str]): List of file extensions to consider.
        converter (PathConverter): Function to convert paths to strings.
    """

    root: pathlib.Path
    look_on: Path | None = None
    exclude: Sequence[StrOrPath] = ()
    exts: Sequence[str] = ()
    converter: PathConverter = Path.as_posix
    include: Sequence[StrOrPath] = ()

    def __post_init__(self):
        self.exclude = ("__pycache__", *self.exclude)
        self.exts = (".py", *self.exts)

    @lazyfield
    def excludes(self) -> tuple[tuple[str, ...], tuple[Path, ...]]:
        str_exclude = tuple(filter_isinstance(str, self.exclude))
        path_exclude = tuple(filter_isinstance(pathlib.Path, self.exclude))
        return str_exclude, path_exclude

    @lazyfield
    def includes(self) -> tuple[tuple[str, ...], tuple[Path, ...]]:
        str_include = tuple(filter_isinstance(str, self.include))
        path_include = tuple(filter_isinstance(pathlib.Path, self.include))
        return str_include, path_include

    def should_lookup(self, path: Path) -> bool:
        str_exclude, path_exclude = self.excludes
        str_include, path_include = self.includes
        _, ext = os.path.splitext(path)

        should_look = (
            path.name not in str_exclude
            and path not in path_exclude
            and (not path.is_file() or ext in self.exts)
        )
        if any((str_include, path_include)):
            should_look = should_look and any(
                (
                    not path.is_dir() and path.name in str_include,
                    path in path_include,
                )
            )
        return should_look

    @property
    def target_path(self) -> Path:
        return self.look_on or self.root

    @lazymethod
    def get_resolver(self) -> PathConverter:
        return make_modulename_converter(self.root, self.converter)

    def iterdir(self, path: Path) -> Iterator[Path]:
        stack = [path]
        while stack:
            path = stack.pop()
            if not path.is_dir():
                yield path
                continue
            for item in filter(self.should_lookup, path.iterdir()):
                if item.is_dir():
                    stack.append(item)
                else:
                    yield item


def sort_files_by_dependency(
    modnames: list[str], files: list[Path]
) -> Iterator[Path]:
    """
    Sorts Python files by their dependencies using AST analysis.

    Args:
        modnames (List[str]): List of module names corresponding to `files`.
        files (List[Path]): List of Path objects representing Python files to sort.

    Returns:
        List[Path]: Sorted list of Python files by dependency, least dependent first.
    """
    # Step 1: Parse each file and extract dependencies
    dependencies = defaultdict(set)
    for modname, file in zip(modnames, files):
        try:
            with file.open("r", encoding="utf-8") as f:
                module_ast = ast.parse(f.read(), filename=str(file))
                for node in ast.walk(module_ast):
                    if isinstance(node, ast.Import):
                        for alias in node.names:
                            if alias.name != "__future__":
                                dependencies[modname].add(alias.name)
                    elif isinstance(node, ast.ImportFrom):
                        if node.module:
                            dependencies[modname].add(node.module)
        except SyntaxError:
            # Handle syntax errors in the files gracefully
            warnings.warn(f"SyntaxError: Unable to parse file {file}")
            continue

    # Step 2: Perform topological sorting based on dependencies
    sorted_files = []
    visited = set()

    def dfs(modname: str):
        if modname in visited:
            return
        visited.add(modname)
        for dep in dependencies[modname]:
            dfs(dep)
        sorted_files.append(modname)

    for modname in modnames:
        dfs(modname)

    # Map sorted module names back to files
    for modname in sorted_files:
        for file, mod in zip(files, modnames):
            if mod == modname:
                yield file


def smart_import(modfullname: str, resolver: PathConverter) -> ModuleType:
    """
    Smart import function that avoids duplicate imports by leveraging sys.modules.

    Args:
        modfullname (str): Full module name to import.
        resolver (Callable[[Path], str]): Function to resolve module paths to full names.

    Returns:
        ModuleType: Imported module.
    """
    *_, modname = modfullname.rsplit(".", 1)
    if modname in sys.modules:
        return sys.modules[modname]

    main = sys.modules.get("__main__")

    main_modname = ""
    if main and (main_file := getattr(main, "__file__", None)):
        main_modname = resolver(Path(main_file))

    if main and main_modname == modname:
        return main
    return importlib.import_module(modfullname)


def ordered_mod(module: ModuleType):
    """Return a list of (name, obj) pairs sorted by place in module source."""
    modlines, _ = getsourcelines(module)
    objects = []
    later = []

    for name, obj in module.__dict__.items():
        try:
            source_lines, _ = getsourcelines(obj)
            start_line = source_lines[0]
            if start_line in modlines:
                linepos = modlines.index(start_line)
                objects.append((name, obj, linepos))
            else:
                later.append((name, obj))
        except (OSError, TypeError):
            later.append((name, obj))
    objects.sort(key=lambda x: x[2])  # Sort by start line number
    return [(name, obj) for name, obj, _ in objects] + later



import pytest

from escudeiro.url.netloc import Netloc


def test_netloc():  # sourcery skip: extract-duplicate-method
    # Test basic netloc
    netloc = Netloc("www.example.com")
    assert netloc.encode() == "www.example.com"
    assert netloc.host == "www.example.com"
    assert netloc.username is None
    assert netloc.password is None
    assert netloc.port is None

    # Test netloc with port
    netloc = Netloc("www.example.com:8080")
    assert netloc.encode() == "www.example.com:8080"
    assert netloc.host == "www.example.com"
    assert netloc.username is None
    assert netloc.password is None
    assert netloc.port == 8080

    # Test netloc with username and password
    netloc = Netloc("username:password@www.example.com")
    assert netloc.encode() == "username:password@www.example.com"
    assert netloc.host == "www.example.com"
    assert netloc.username == "username"
    assert netloc.password == "password"
    assert netloc.port is None

    # Test netloc from args
    netloc = Netloc.from_args(
        host="www.example.com",
        username="username",
        password="password",
    )
    assert netloc == "username:password@www.example.com"
    assert netloc.host == "www.example.com"
    assert netloc.username == "username"
    assert netloc.password == "password"
    assert netloc.port is None

    # Test merging works

    instance = Netloc("www.example.com")

    assert (
        instance.merge(
            Netloc.from_args(
                host="www.example2.com",
                username="username",
                password="password",
            )
        )
        == "username:password@www.example2.com"
    )
    assert (
        instance.merge_left(
            Netloc.from_args(
                host="www.example2.com",
                username="username",
                password="password",
            )
        )
        == "username:password@www.example.com"
    )


def test_netloc_does_bounds_check_for_port():
    with pytest.raises(ValueError) as exc_info:
        _ = Netloc("www.example.com:66000").encode()

    assert (
        exc_info.value.args[0] == "Invalid port received while parsing netloc."
    )

    netloc = Netloc("example.com")
    with pytest.raises(ValueError) as exc_info:
        netloc.port = 66000

    assert exc_info.value.args[0] == "Received invalid port value."

    netloc = Netloc("example.com")
    with pytest.raises(ValueError) as exc_info:
        netloc.port = -1

    assert exc_info.value.args[0] == "Received invalid port value."



# pyright: reportPrivateUsage=false
import asyncio
import contextlib
import threading
import time
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    nullcontext,
)
from types import TracebackType
from typing import Any, Self, final, override
from unittest.mock import AsyncMock, Mock

from escudeiro.lazyfields import (
    _UNSET,
    ALazyContainer,
    AsyncLazyField,
    LazyContainer,
    LazyField,
    asynclazyfield,
    lazyfield,
    mark_class,
)


@final
class MockContextManager(AbstractContextManager, AbstractAsyncContextManager):
    """Reusable mock context manager for both sync and async contexts."""

    def __init__(self) -> None:
        super().__init__()
        self.call_count = 0
        self.enter_calls = 0
        self.exit_calls = 0

    def __call__(self) -> Self:
        return self

    @override
    def __enter__(self):
        super().__enter__()
        self.call_count += 1
        self.enter_calls += 1
        return self

    @override
    async def __aenter__(self):
        await super().__aenter__()
        self.call_count += 1
        self.enter_calls += 1
        return self

    @override
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        self.exit_calls += 1
        return

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ):
        self.exit_calls += 1
        return


@final
class LazyFieldTestFixture:
    """Test fixture for LazyField tests to avoid class creation in each test."""

    def __init__(
        self,
        lock_factory: Callable[
            [], AbstractContextManager
        ] = contextlib.nullcontext,
        field_value: Any = "test",
    ):
        self.mock = Mock()
        self.lock_factory = lock_factory
        self.field_value = field_value
        self.sample_class = self.create_sample_class()

    def create_sample_class(self):
        mock = self.mock
        field_value = self.field_value
        lock_factory = self.lock_factory

        @mark_class(lock_factory)
        class Sample:
            @lazyfield(lock_factory=lock_factory)
            def field(self):
                mock()
                return field_value

        return Sample

    def create_instance(self):
        return self.sample_class()

    def get_container(self, instance: Any) -> LazyContainer:
        """Get the LazyContainer from an instance."""
        return getattr(instance, LazyField.make_private("field"))

    def assert_container_exists(self, instance: Any):
        """Assert that the container exists on the instance."""
        assert hasattr(instance, LazyField.make_private("field"))

    def assert_container_content(self, instance: Any, expected_content: Any):
        """Assert that the container has the expected content."""
        container = self.get_container(instance)
        assert container.content == expected_content

    def assert_container_content_is_unset(self, instance: Any):
        """Assert that the container content is _UNSET."""
        container = self.get_container(instance)
        assert container.content is _UNSET


def test_lazyfield_freezes_value():
    """Test that LazyField freezes the value for each instance."""

    class Sample:
        @LazyField
        def frozen(self) -> list:
            return []

    sample = Sample()
    sample_2 = Sample()

    assert sample.frozen is sample.frozen
    assert sample_2.frozen is not sample.frozen
    assert isinstance(sample.frozen, list)


def test_lazyfield_creates_hidden_container():
    """Test that LazyField creates a hidden container attribute."""
    fixture = LazyFieldTestFixture()
    sample = fixture.create_instance()

    # Access to trigger initialization
    sample.field

    fixture.assert_container_exists(sample)
    fixture.assert_container_content(sample, "test")

    # Test the lock type
    container = fixture.get_container(sample)
    assert isinstance(container.lock, nullcontext)


def test_lazyfield_calls_context_on_get():
    """Test that LazyField calls the context manager on get."""
    mock_cm = MockContextManager()
    fixture = LazyFieldTestFixture(lock_factory=mock_cm)

    sample = fixture.create_instance()

    # First access triggers initialization and calls enter
    sample.field
    # Additional accesses should also call enter
    sample.field
    sample.field

    # Create another instance to verify per-instance behavior
    sample_2 = fixture.create_instance()
    sample_2.field

    # Each get should call enter
    assert mock_cm.enter_calls > 0, (
        "Lazyfield should call enter on each __get__"
    )


def test_lazyfield_calls_context_on_write():
    """Test that LazyField calls the context manager on write and delete."""
    mock_cm = MockContextManager()
    fixture = LazyFieldTestFixture(lock_factory=mock_cm)

    sample = fixture.create_instance()

    initial_calls = mock_cm.enter_calls
    sample.field = "World"
    assert mock_cm.enter_calls > initial_calls, "Should call enter on __set__"

    initial_calls = mock_cm.enter_calls
    del sample.field
    assert mock_cm.enter_calls > initial_calls, (
        "Should call enter on __delete__"
    )


def test_lazyfield_sets_container_content_to_unset_on_delete():
    """Test that LazyField sets container content to _UNSET on delete."""
    fixture = LazyFieldTestFixture()
    sample = fixture.create_instance()

    # Initialize
    sample.field
    fixture.assert_container_content(sample, "test")

    # Delete
    del sample.field
    fixture.assert_container_content_is_unset(sample)


def test_get_after_del_resets_function_state():
    """Test that getting after deleting resets the function state."""
    fixture = LazyFieldTestFixture()
    sample = fixture.create_instance()

    # First access
    sample.field
    # Second access shouldn't call the function again
    sample.field
    assert fixture.mock.call_count == 1

    # Delete and access again
    del sample.field
    sample.field

    # Function should be called again
    assert fixture.mock.call_count == 2


def test_lazyfield_is_only_computed_on_call():
    """Test that LazyField is only computed when accessed."""
    fixture = LazyFieldTestFixture()
    sample = fixture.create_instance()

    # Should not be called until accessed
    assert fixture.mock.call_count == 0

    # Access triggers computation
    sample.field
    assert fixture.mock.call_count == 1


def test_lazyfield_set_forces_value_into_field():
    """Test that setting LazyField forces a value into the field."""
    sentinel = object()
    fixture = LazyFieldTestFixture()
    sample = fixture.create_instance()

    # Set value
    sample.field = sentinel

    # Value should be set without calling function
    assert sample.field is sentinel
    assert fixture.mock.call_count == 0

    # Container should exist
    fixture.assert_container_exists(sample)


def test_lazyfield_with_different_lock_factories():
    """Test LazyField with different lock factories."""
    lock_types = [nullcontext, MockContextManager()]

    for lock in lock_types:
        fixture = LazyFieldTestFixture(lock_factory=lock)
        sample = fixture.create_instance()

        sample.field
        container = fixture.get_container(sample)

        if isinstance(lock, type) and issubclass(lock, nullcontext):  # pyright: ignore[reportUnnecessaryIsInstance]
            assert isinstance(container.lock, nullcontext)
        else:
            assert container.lock is lock


def test_lazyfield_with_inheritance():
    """Test that LazyField works with inheritance."""

    class Base:
        @lazyfield
        def field(self):
            return "base"

    class Child(Base):
        @lazyfield
        def field(self):
            return "child"

    base = Base()
    child = Child()

    assert base.field == "base"
    assert child.field == "child"


def test_lazyfield_with_property_descriptor_protocol():
    """Test that LazyField follows the property descriptor protocol."""

    class Sample:
        @lazyfield
        def field(self):
            return "test"

    # Descriptor protocol should work on class
    descriptor = Sample.__dict__["field"]
    assert isinstance(descriptor, LazyField)

    # __get__ with instance=None should return the descriptor itself
    assert descriptor.__get__(None, Sample) is descriptor


def test_lazyfield_concurrent_access():
    """Test that AsyncLazyField handles concurrent access correctly."""

    @final
    @mark_class(threading.Lock)
    class SlowSample:
        call_count = 0

        @lazyfield(lock_factory=threading.Lock)
        def field(self):
            self.call_count += 1
            time.sleep(0.1)  # Simulate slow computation
            return f"result-{self.call_count}"

    sample = SlowSample()

    # Concurrent access should only compute once
    with ThreadPoolExecutor() as executor:
        results = tuple(executor.map(lambda _: sample.field, range(3)))

    # All results should be the same
    assert all(r == results[0] for r in results), results
    assert sample.call_count == 1


class MockAsyncContextManager(contextlib.AbstractAsyncContextManager):
    """Reusable mock async context manager for testing."""

    def __init__(self) -> None:
        super().__init__()
        self.call_count: int = 0
        self.enter_calls: int = 0
        self.exit_calls: int = 0

    def __call__(self) -> Self:
        return self

    @override
    async def __aenter__(self):
        await super().__aenter__()
        self.call_count += 1
        self.enter_calls += 1
        return self

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ):
        self.exit_calls += 1
        return


@final
class AsyncLazyFieldTestFixture:
    """Test fixture for AsyncLazyField tests to avoid class creation in each test."""

    def __init__(
        self,
        lock_factory: Callable[[], AbstractAsyncContextManager] = nullcontext,
        field_value: Any = "test",
    ):
        self.mock = AsyncMock()
        self.lock_factory = lock_factory or contextlib.nullcontext
        self.field_value = field_value
        self.sample_class = self.create_sample_class()

    def create_sample_class(self):
        mock = self.mock
        field_value = self.field_value
        lock_factory = self.lock_factory

        @mark_class(actx_factory=lock_factory)
        class Sample:
            @asynclazyfield(lock_factory=lock_factory)
            async def field(self):
                await mock()
                return field_value

        return Sample

    def create_instance(self):
        return self.sample_class()

    def get_container(self, instance: Any) -> ALazyContainer:
        """Get the AlazyContainer from an instance."""
        return getattr(instance, AsyncLazyField.make_private("field"))

    def assert_container_exists(self, instance: Any):
        """Assert that the container exists on the instance."""
        assert hasattr(instance, AsyncLazyField.make_private("field"))

    async def assert_container_content(
        self, instance: Any, expected_content: Any
    ):
        """Assert that the container has the expected content."""
        container = self.get_container(instance)
        content = await container.acquire()
        assert content == expected_content

    async def assert_container_content_is_unset(self, instance: Any):
        """Assert that the container content is _UNSET."""
        container = self.get_container(instance)
        content = await container.acquire()
        assert content is _UNSET


async def test_asynclazyfield_freezes_value():
    """Test that AsyncLazyField freezes the value for each instance."""

    class Sample:
        @asynclazyfield
        async def frozen(self) -> list:
            return []

    sample = Sample()
    sample_2 = Sample()

    result1 = await sample.frozen()
    result2 = await sample.frozen()
    result3 = await sample_2.frozen()

    assert result1 is result2
    assert result2 is not result3
    assert isinstance(result1, list)


async def test_asynclazyfield_creates_hidden_container():
    """Test that AsyncLazyField creates a hidden container attribute."""
    fixture = AsyncLazyFieldTestFixture()
    sample = fixture.create_instance()

    # Access to trigger initialization
    await sample.field()

    fixture.assert_container_exists(sample)
    await fixture.assert_container_content(sample, "test")

    # Test the lock type
    container = fixture.get_container(sample)
    assert isinstance(container.lock, contextlib.nullcontext)


async def test_asynclazyfield_calls_context_on_get():
    """Test that AsyncLazyField calls the context manager on get."""
    mock_cm = MockAsyncContextManager()
    fixture = AsyncLazyFieldTestFixture(lock_factory=mock_cm)

    sample = fixture.create_instance()
    # First access triggers initialization and calls enter
    await sample.field()
    initial_enter_calls = mock_cm.enter_calls

    # Additional accesses should also call enter
    await sample.field()
    await sample.field()

    # Create another instance to verify per-instance behavior
    sample_2 = fixture.create_instance()
    await sample_2.field()

    # Each get should call enter
    assert mock_cm.enter_calls > initial_enter_calls, (
        "AsyncLazyField should call enter on each __get__"
    )


async def test_asynclazyfield_is_only_computed_on_call():
    """Test that AsyncLazyField is only computed when accessed."""
    fixture = AsyncLazyFieldTestFixture()
    sample = fixture.create_instance()

    # Should not be called until accessed
    assert fixture.mock.await_count == 0

    # Access triggers computation
    await sample.field()
    assert fixture.mock.await_count == 1


async def test_asynclazyfield_reset_with_none():
    """Test that resetting AsyncLazyField with None deletes the content."""
    fixture = AsyncLazyFieldTestFixture()
    sample = fixture.create_instance()

    # Initialize
    await sample.field()
    await fixture.assert_container_content(sample, "test")

    # Reset with None should delete
    descriptor = type(sample).__dict__["field"]
    await descriptor.reset(sample)

    await fixture.assert_container_content_is_unset(sample)


async def test_asynclazyfield_get_after_reset():
    """Test that getting after resetting recomputes the value."""
    fixture = AsyncLazyFieldTestFixture()
    sample = fixture.create_instance()

    # First access
    await sample.field()
    # Second access shouldn't call the function again
    await sample.field()
    assert fixture.mock.await_count == 1

    # Reset and access again
    descriptor = sample.__class__.__dict__["field"]
    await descriptor.reset(sample)
    await sample.field()

    # Function should be called again
    assert fixture.mock.await_count == 2


async def test_asynclazyfield_reset_with_value():
    """Test that resetting AsyncLazyField with a value sets the content."""
    fixture = AsyncLazyFieldTestFixture()
    sample = fixture.create_instance()

    # Initialize
    await sample.field()

    # Reset with value
    descriptor = sample.__class__.__dict__["field"]
    await descriptor.reset(sample, "new_value")

    # Should have new value without calling function
    result = await sample.field()
    assert result == "new_value"
    assert fixture.mock.await_count == 1  # Only the initial call


async def test_asynclazyfield_reset_on_uninitialized():
    """Test that resetting an uninitialized AsyncLazyField works correctly."""
    fixture = AsyncLazyFieldTestFixture()
    sample = fixture.create_instance()

    # Reset without initialization
    descriptor = sample.__class__.__dict__["field"]
    await descriptor.reset(sample, "direct_value")

    # Should have the value without calling function
    result = await sample.field()
    assert result == "direct_value"
    assert fixture.mock.await_count == 0


async def test_asynclazyfield_concurrent_access():
    """Test that AsyncLazyField handles concurrent access correctly."""

    @final
    @mark_class(actx_factory=asyncio.Lock)
    class SlowSample:
        call_count = 0

        @asynclazyfield(lock_factory=asyncio.Lock)
        async def field(self):
            self.call_count += 1
            await asyncio.sleep(0.1)  # Simulate slow computation
            return f"result-{self.call_count}"

    sample = SlowSample()

    # Concurrent access should only compute once
    results = await asyncio.gather(
        sample.field(), sample.field(), sample.field()
    )

    # All results should be the same
    assert all(r == results[0] for r in results), results
    assert sample.call_count == 1


async def test_asynclazyfield_with_different_lock_factories():
    """Test AsyncLazyField with different lock factories."""
    results = []

    class CustomLockManager(MockAsyncContextManager):
        @override
        async def __aenter__(self):
            result = await super().__aenter__()
            results.append("lock entered")
            return result

    lock_factory = CustomLockManager()
    fixture = AsyncLazyFieldTestFixture(lock_factory=lock_factory)
    sample = fixture.create_instance()

    await sample.field()
    assert "lock entered" in results
    assert lock_factory.enter_calls > 0


async def test_asynclazyfield_with_inheritance():
    """Test that AsyncLazyField works with inheritance."""

    class Base:
        @asynclazyfield
        async def field(self):
            return "base"

    class Child(Base):
        @asynclazyfield
        async def field(self):
            return "child"

    base = Base()
    child = Child()

    assert await base.field() == "base"
    assert await child.field() == "child"


async def test_asynclazyfield_descriptor_protocol():
    """Test that AsyncLazyField follows the descriptor protocol."""

    class Sample:
        @asynclazyfield
        async def field(self):
            return "test"

    # Descriptor protocol should work on class
    descriptor = Sample.__dict__["field"]
    assert isinstance(descriptor, AsyncLazyField)

    # __get__ with instance=None should return the descriptor itself
    assert descriptor.__get__(None, Sample) is descriptor


async def test_alazycontainer_operations():
    """Test AlazyContainer operations directly."""
    lock = MockAsyncContextManager()
    container = ALazyContainer("initial", lock)

    # Test acquire
    content = await container.acquire()
    assert content == "initial"
    assert lock.enter_calls == 1
    assert lock.exit_calls == 1

    # Test put
    await container.put("updated")
    content = await container.acquire()
    assert content == "updated"
    assert lock.enter_calls == 3
    assert lock.exit_calls == 3

    # Test delete
    await container.delete()
    content = await container.acquire()
    assert content is _UNSET
    assert lock.enter_calls == 5
    assert lock.exit_calls == 5


async def test_asynclazyfield_decorated_methods():
    """Test different ways of decorating with asynclazyfield."""

    # Test with direct decoration
    class Sample1:
        @asynclazyfield
        async def method1(self):
            return "method1"

    # Test with function call decoration
    class Sample2:
        @asynclazyfield()
        async def method2(self):
            return "method2"

    # Test with custom lock factory
    lock = MockAsyncContextManager()

    @mark_class(actx_factory=lock)
    class Sample3:
        @asynclazyfield(lock_factory=lambda: lock)
        async def method3(self):
            return "method3"

    s1 = Sample1()
    s2 = Sample2()
    s3 = Sample3()

    assert await s1.method1() == "method1"
    assert await s2.method2() == "method2"
    assert await s3.method3() == "method3"
    assert lock.enter_calls > 0


class TestComplexUseCases:
    """Test complex use cases with AsyncLazyField."""

    async def test_multiple_asynclazyfields_same_class(self):
        """Test multiple AsyncLazyFields on the same class."""
        mock1 = AsyncMock(return_value="value1")
        mock2 = AsyncMock(return_value="value2")

        class MultiFields:
            @asynclazyfield
            async def field1(self):
                await mock1()
                return "value1"

            @asynclazyfield
            async def field2(self):
                await mock2()
                return "value2"

        instance = MultiFields()

        assert await instance.field1() == "value1"
        assert await instance.field2() == "value2"

        # Verify each function was called exactly once
        assert mock1.await_count == 1
        assert mock2.await_count == 1

        # Access again - should use cached values
        assert await instance.field1() == "value1"
        assert await instance.field2() == "value2"

        assert mock1.await_count == 1
        assert mock2.await_count == 1

    async def test_asynclazyfield_depending_on_another(self):
        """Test an AsyncLazyField that depends on another AsyncLazyField."""
        first_mock = AsyncMock(return_value="first")
        second_mock = AsyncMock()

        class DependentFields:
            @asynclazyfield
            async def first(self):
                await first_mock()
                return "first"

            @asynclazyfield
            async def second(self):
                first_value = await self.first()
                await second_mock(first_value)
                return f"second-{first_value}"

        instance = DependentFields()

        # Access the dependent field
        result = await instance.second()

        assert result == "second-first"
        assert first_mock.await_count == 1
        second_mock.assert_called_once_with("first")

        # Access again - both should use cached values
        result = await instance.second()
        assert result == "second-first"
        assert first_mock.await_count == 1
        assert second_mock.call_count == 1



import pytest


@pytest.fixture(autouse=True)
def no_env_vars(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv("CONFIG_ENV", raising=False)
    monkeypatch.delenv("CONFIG_DOTFILE", raising=False)
    monkeypatch.delenv("NAME", raising=False)
    yield



from collections.abc import Generator, Sequence
from typing import Any, override

from escudeiro.config.adapter.interface import FieldResolverStrategy
from escudeiro.config.interface import MISSING
from escudeiro.data import data, get_fields
from escudeiro.data.field_ import Field
from escudeiro.data.utils.typedef import MISSING as NOTHING


@data
class SquireResolverStrategy(FieldResolverStrategy[Field]):
    field: Field

    @override
    def cast(self) -> type:
        return self.field.declared_type

    @override
    def names(self) -> Sequence[str]:
        return self.field.name, self.field.alias

    @override
    def init_name(self) -> str:
        return self.field.alias or self.field.name

    @override
    def default(self) -> Any | type[MISSING]:
        default = self.field.default
        if default is NOTHING:
            return MISSING
        return default() if self.field.has_default_factory else default

    @override
    @classmethod
    def iterfield(cls, config_class: type) -> Generator[Field, Any, Any]:
        _ = cls
        yield from get_fields(config_class).values()



from collections.abc import Sequence
from typing import Protocol, final, override

from escudeiro.misc import to_camel

template = (
    '"title": "{title}"',
    '"type": "{type}"',
)


class HasStr(Protocol):
    @override
    def __str__(self) -> str: ...


class HasToString(HasStr, Protocol):
    def to_string(self) -> str: ...


type PropsType = dict[str, HasStr] | HasStr


def str_cast(string: str):
    return f'"{string}"'


@final
class DictSchema:
    __slots__ = ("type_", "extras")

    def __init__(self, type_: str, **extras: PropsType) -> None:
        self.type_ = type_
        self.extras = extras

    @override
    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        return (
            "{"
            + ",".join(
                item
                for item in (f'"type": "{self.type_}"', self.build_extras())
                if item
            )
            + "}"
        )

    def build_extras(self) -> str:
        if not self.extras:
            return ""
        lines = [
            f'"{to_camel(name)}":'
            + (
                " {"
                + ", ".join(
                    f'"{to_camel(key)}": {value!s}'
                    for key, value in props.items()
                )
                + "}"
                if isinstance(props, dict)
                else str(props)
            )
            for name, props in self.extras.items()
        ]
        return ", ".join(lines)


@final
class ListSchema:
    __slots__ = (
        "type_",
        "items",
    )

    def __init__(self, type_: str, *items: HasStr) -> None:
        self.type_ = type_
        self.items = items

    @override
    def __str__(self):
        return self.to_string()

    def to_string(self):
        items = ", ".join(str(item) for item in self.items)
        return f'{{"{self.type_}": [{items}]}}'


@final
class Items:
    __slots__ = ("items",)

    def __init__(self, *items: HasStr) -> None:
        self.items = items

    @override
    def __str__(self):
        return self.to_string()

    def to_string(self):
        items = ", ".join(str(item) for item in self.items)
        return f"[{items}]"


@final
class RefSchema:
    __slots__ = ("title",)

    def __init__(self, title: str) -> None:
        self.title = title

    def to_string(self) -> str:
        return f'{{"$ref": "#/components/schemas/{self.title}"}}'


@final
class Schema:
    __slots__ = ("title", "type_", "required", "extras", "nullable")

    def __init__(
        self,
        title: str,
        type_: str,
        required: Sequence[str] = (),
        nullable: bool = False,
        **extras: PropsType,
    ) -> None:
        self.title = title
        self.type_ = type_
        self.required = required
        self.nullable = nullable
        self.extras = extras

    @override
    def __str__(self):
        return self.to_string()

    def to_string(self) -> str:
        extras = self.build_extras()
        required = self.build_required()
        return (
            "{"
            + ", ".join(
                item
                for item in (
                    f'"title": "{self.title}"',
                    f'"type": "{self.type_}"',
                    required,
                    extras,
                    f'"nullable": {str(self.nullable)}',
                )
                if item
            )
            + "}"
        )

    def build_extras(self) -> str:
        if not self.extras:
            return ""
        lines = []
        for name, props in self.extras.items():
            props_str = self._make_prop_str(props)
            lines.append(f'"{to_camel(name)}": {props_str}')
        return ", ".join(lines)

    def _make_prop_str(self, props: PropsType):
        if not isinstance(props, dict):
            return str(props)
        props_str = ",".join(
            f'"{to_camel(key)}": {value!s}' for key, value in props.items()
        )
        return f"{{{props_str}}}"

    def build_required(self) -> str:
        return (
            f""""required": [{",".join(f'"{val}"' for val in self.required)}]"""
            if self.required
            else ""
        )



import pathlib

import pytest


@pytest.fixture
def sample_root_path() -> pathlib.Path:
    return pathlib.Path(__file__).parent.parent


@pytest.fixture
def mock_module_path(sample_root_path: pathlib.Path) -> pathlib.Path:
    return sample_root_path / "autodiscovery/modules"


@pytest.fixture
def sample_files(mock_module_path: pathlib.Path) -> list[pathlib.Path]:
    return list((mock_module_path).glob("**/*.py"))



from __future__ import annotations

import os
from pathlib import Path
from types import TracebackType
from typing import Self

from escudeiro.data import data
from escudeiro.ds.filetree.virtual import VirtualFileTree
from escudeiro.escudeiro_pyrs import filetree
from escudeiro.exc import FailedFileOperation, InvalidPath
from escudeiro.lazyfields import lazyfield


@data(frozen=False)
class FileTree:
    base_dir: Path

    @lazyfield
    def root(self):
        return filetree.FsNode(self.base_dir.name)

    @lazyfield
    def virtual(self):
        return VirtualFileTree(self.root)

    def write(self):
        self.write_tree(self.base_dir, self.root)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        *exc_info: *tuple[
            type[BaseException] | None,
            BaseException | None,
            TracebackType | None,
        ],
    ):
        if any(exc_info):
            _, exc, _ = exc_info

            raise FailedFileOperation(
                "unable to write to disk after exception", exc
            ) from exc
        self.write()

    def write_tree(self, path: Path, node: filetree.FsNode) -> None:
        stack = [(path, node)]

        while stack:
            curpath, curnode = stack.pop()
            if curnode.is_file():
                # file
                self._write_file(curpath, curnode)
            else:
                # folder
                self._write_folder(curpath)
                for node in curnode.children:
                    stack.append((curpath / node.name, node))

    def _write_file(self, path: Path, node: filetree.FsNode):
        if node.content is None:
            raise ValueError(f"Expected file node, got folder at {path}")
        if path.exists() and path.is_dir():
            path.rmdir()
        elif path.exists() and path.is_file():
            path.unlink(missing_ok=True)
        else:
            path.touch()
        with open(path, "wb") as stream:
            _ = stream.write(node.content)

    def _write_folder(self, path: Path):
        if path.exists():
            if path.is_file():
                path.unlink(missing_ok=True)
        else:
            path.mkdir(parents=True)

    def merge(self, tree: FileTree | VirtualFileTree, *path: str) -> None:
        if isinstance(tree, FileTree):
            self.virtual.merge(tree.virtual, *path)
        else:
            self.virtual.merge(tree, *path)

    def validate(self) -> None:
        """
        Validate that the virtual file tree can be safely written to the base_dir in the actual filesystem.
        This includes checking for:
            - Path conflicts (file where folder is expected, etc.)
            - Missing intermediate directories
            - Permissions (read/write) — if needed
        Raises:
            FailedFileOperation: If any part of the tree would fail to be written due to a conflict.
        """
        stack = [(self.base_dir, self.virtual.root)]

        while stack:
            curpath, curnode = stack.pop()

            # Check for invalid file/folder conflicts
            if curpath.exists():
                if not curnode.is_file() and not curpath.is_dir():
                    raise InvalidPath(
                        f"Expected folder at {curpath}, but found file or other"
                    )
                elif curnode.is_file() and not curpath.is_file():
                    raise InvalidPath(
                        f"Expected file at {curpath}, but found directory or other"
                    )

            # Traverse children if directory
            if not curnode.is_file():
                for child in curnode.children:
                    stack.append((curpath / child.name, child))

            # Optional: verify write permission
            if curpath.exists():
                if not os.access(curpath, os.W_OK):
                    raise FailedFileOperation(
                        f"No write permission for {curpath}"
                    )
            else:
                # Check parent directory for writability
                parent = curpath.parent
                if not parent.exists() or not os.access(parent, os.W_OK):
                    raise FailedFileOperation(
                        f"Cannot write to {parent} (nonexistent or unwritable)"
                    )



import asyncio
import contextlib
import threading

from escudeiro.context.interfaces import Adapter, AsyncAdapter
from escudeiro.data import data
from escudeiro.lazyfields import asynclazyfield, dellazy, lazyfield, mark_class


@mark_class(threading.RLock)
@data(frozen=False)
class Context[T]:
    adapter: Adapter[T]

    @lazyfield
    def stack(self) -> int:
        return 0

    @lazyfield
    def _lock(self) -> threading.Lock:
        return threading.Lock()

    @lazyfield
    def client(self) -> T:
        """
        Returns the current resource being used by the context.
        Acquires a new resource if the current one is closed or doesn't exist.
        """
        return self.adapter.new()

    def is_active(self) -> bool:
        """
        Returns whether the context is currently in use.
        """
        return self.stack > 0

    def acquire(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        with self._lock:
            if self.adapter.is_closed(self.client):
                dellazy(self, type(self).client)
            self.stack += 1
            return self.client

    def release(self):
        """
        Releases the current resource if the stack count is 1,
        and decreases the stack count.
        """
        if self.stack == 1:
            with self._lock:
                self.adapter.release(self.client)
                dellazy(self, type(self).client)
        self.stack -= 1

    @contextlib.contextmanager
    def open(self):
        """
        A context manager that acquires and releases resources without returning it.
        """
        with self:
            yield

    @contextlib.contextmanager
    def begin(self):
        """
        A context manager that acquires and releases resources and returns it.
        """
        with self as client:
            yield client

    def __enter__(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        return self.acquire()

    def __exit__(self, *_):
        """
        Releases the current resource if the stack count is 1,
        and decreases the stack count.
        """
        self.release()


@data(frozen=False)
class AsyncContext[T]:
    adapter: AsyncAdapter[T]

    @lazyfield
    def stack(self):
        return 0

    @lazyfield
    def lock(self):
        return asyncio.Lock()

    def is_active(self) -> bool:
        """
        Returns whether the context is currently in use.
        """
        return self.stack > 0

    @asynclazyfield
    async def client(self) -> T:
        """
        Returns the current resource being used by the context.
        Acquires a new resource if the current one is closed or doesn't exist.
        """
        return await self.adapter.new()

    async def acquire(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        async with self.lock:
            client = await self.client()
            self.stack += 1
            return client

    async def release(self):
        """
        Releases the current resource if the stack count is 1,
        and decreases the stack count.
        """
        async with self.lock:
            if self.stack == 1:
                await self.adapter.release(await self.client())
                await dellazy(self, type(self).client)
            self.stack -= 1

    @contextlib.asynccontextmanager
    async def open(self):
        """
        An async context manager that acquires and releases resources without returning it.
        """
        async with self:
            yield

    @contextlib.asynccontextmanager
    async def begin(self):
        """
        An async context manager that acquires and releases resources and returns it.
        """
        async with self as client:
            yield client

    async def __aenter__(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        return await self.acquire()

    async def __aexit__(self, *_):
        """
        Releases the current resource if the stack count is 1,
        and decreases the stack count.
        """
        await self.release()



from escudeiro.url.fragment import Fragment


def test_fragment():
    # Test creating a fragment from a string
    fragment = Fragment("test")
    assert fragment.encode() == "test"

    # Test encoding a fragment with special characters
    fragment = Fragment("test#test")
    assert fragment.encode() == "test%23test"

    # Test encoding with encoded does not double encodes
    fragment = Fragment("test%23test")
    assert fragment.encode() == "test%23test"

    # Test updating a fragment
    _ = fragment.set("new")
    assert fragment.encode() == "new"
